# -*- coding: utf-8 -*-
'''
 brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
 Copyright (C) 2018-2019,  F.P.A. Vogt
 
 -----------------------------------------------------------------------------------------
 
 This file contains general tools for the brutifus routines to fit the stellar continuum 
 and the emission lines in an IFU data cube.

 Created November 2018, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au
'''
# ----------------------------------------------------------------------------------------

import os
import numpy as np
import signal 

from astropy.constants import c
from astropy.io import fits

from .brutifus_version import __version__ as version
from . import brutifus_metadata as bifus_m
 
# ----------------------------------------------------------------------------------------       
def extract_cube(fn,inst):
   ''' Extracts the data and error associated with a given datacube.
   
   :param fn: relative path to file
   :type fn: string
   :param inst: Name of the instrument that took the data
   :type inst: string
   
   :return: [[lams,data,error], [header0, header_data, header_error]]
   :rtype: list
   
   '''
   
   if not(os.path.isfile(fn)):
      raise Exception('File not found: %s' % (fn))
   
   if not(inst in bifus_m.ffmt.keys()):
      raise Exception('Instrument not supported: %s' % (inst))
   
   # Open the FITS file, and extract the info I need
   hdu = fits.open(fn)
   
   if inst == 'MUSE':
      header0 = hdu[0].header
   else:
      header0 = None
      
   data = hdu[bifus_m.ffmt[inst]['data']].data
   header_data = hdu[bifus_m.ffmt[inst]['data']].header
   error = hdu[bifus_m.ffmt[inst]['var']].data
   header_error = hdu[bifus_m.ffmt[inst]['var']].header
   hdu.close()
    
   # Build the wavelength array - REST frame !
   lams = np.arange(0, header_data['NAXIS3'],1) * \
            header_data['CD3_3'] + header_data['CRVAL3']
   
   
   return [[lams,data,error], [header0, header_data, header_error]]
 
# ----------------------------------------------------------------------------------------  
def nearest_2dpoint(point, points):
   ''' Returns the nearest neighbor from a bung of points, and the distance
   
   :param point: 1x2 array with x, y coord
   :type point: numpy array
   :param points: Nx2 array with x, y coords
   :type points: numpy array
   
   :return: the min distance, the delta-x and delta-y, and the closest point
   :rtype: list
   
   '''
   
   dist_2 = np.sum((points - point)**2, axis = 1)
   min_id = np.argmin(dist_2)
   return (np.sqrt(dist_2[min_id]), points[min_id] - point, points[min_id])

# ----------------------------------------------------------------------------------------       
def init_worker():
   ''' Handles KeyboardInterrupt during multiprocessing.
   
   .. note:: See https://noswap.com/blog/python-multiprocessing-keyboardinterrupt
   
   '''
   signal.signal(signal.SIGINT, signal.SIG_IGN)
# ----------------------------------------------------------------------------------------      
  
def hdu_add_brutifus(hdu,procstep):
   ''' Adds dedicated brutifus keywords to a FITS file header.
    
   :param hdu: The destination hdu to which the brutifus keywords must be added.
   :param procstep: The name of the processing step creating the FITS file.
   :type procstep: str
   
   :return: The newheader with brutifus info included.  
    
   '''
    
   hdu.header['BRUTIFUS'] = (version, 'brutifus version')
   hdu.header['B_STEP'] = (procstep, 'brutifus processing step')
    
   return hdu
    
# ----------------------------------------------------------------------------------------      
  
def hdu_add_wcs(newhdu,refheader):
   ''' Adds the WCS coordinates from a reference header to a new hdu.
    
   :param newheader: The destination hdu to which the WCS keywords must be added.
   :param refheader: The reference header, from which to transer the WCS keywords.
    
   :return: The new hdu with WCS info included. 
     
   .. note:: Keywords transfered are  'CRPIX1', 'CD1_1', 'CTYPE1', 'CUNIT1', 'CRPIX2', 
             'CD2_2', 'CTYPE2', 'CUNIT2', 'CD1_2', 'CD2_1', 'CRVAL1' and 'CRVAL2'.
   '''
    
   newhdu.header['CRPIX1'] = refheader['CRPIX1']
   newhdu.header['CD1_1'] = refheader['CD1_1']
   newhdu.header['CTYPE1'] = refheader['CTYPE1']
   newhdu.header['CUNIT1'] = refheader['CUNIT1']
   newhdu.header['CRPIX2'] = refheader['CRPIX2']
   newhdu.header['CD2_2'] = refheader['CD2_2']
   newhdu.header['CTYPE2'] = refheader['CTYPE2']
   newhdu.header['CUNIT2'] = refheader['CUNIT2']
   newhdu.header['CD1_2'] = refheader['CD1_2']
   newhdu.header['CD2_1'] = refheader['CD2_1']
   newhdu.header['CRVAL1'] = refheader['CRVAL1']
   newhdu.header['CRVAL2'] = refheader['CRVAL2']
   
   return newhdu
# ----------------------------------------------------------------------------------------      
    
def hdu_add_lams(newhdu,refheader):
   ''' Adds the wavelength information from a reference header to a new hdu.
    
   :param newhdu: The destination hdu to which the wavelength keywords must be added.
   :param refheader: The reference header, from which to transer the wavelength keywords.
    
   :return: FITS HDU, the newheader with wavelength info included. 
     
   .. note:: Keywords transfered are 'CTYPE3', 'CUNIT3', 'CD3_3', 'CRPIX3', 'CRVAL3', 
             'CD1_3', 'CD2_3', 'CD3_1' and 'CD3_2'.
   '''
    
   newhdu.header['CTYPE3'] = refheader['CTYPE3'] 
   newhdu.header['CUNIT3'] = refheader['CUNIT3']   
   newhdu.header['CD3_3'] = refheader['CD3_3']
   newhdu.header['CRPIX3'] = refheader['CRPIX3']
   newhdu.header['CRVAL3'] = refheader['CRVAL3']
   newhdu.header['CD1_3'] = refheader['CD1_3']
   newhdu.header['CD2_3'] = refheader['CD2_3']
   newhdu.header['CD3_1'] = refheader['CD3_1']
   newhdu.header['CD3_2'] = refheader['CD3_2']
 
   return newhdu    
 
# ----------------------------------------------------------------------------------------
def inst_resolution(inst = 'MUSE', get_ff = False, show_plot = False):
   ''' Returns the functional resolution of an instrument as a function of the wavelength.
    
   Returns a callable function of the wavelength (in Angstroem !).
    
   :param inst: The name tag referring to a given instrument.
   :type inst: str
   :param get_ff: Whether to recompute the given function from a reference dataset or not.
                  Only valid with inst = 'MUSE'.
   :type get_ff: bool
   :param show_plot: Whether to make a plot of the function.
   :type show_plot: bool
   
   :return: A function that takes a float (lambda in Angstroem), and returns the 
            corresponding value of the chosen instrument resolution.
   :rtype: func
    
   .. note:: Supported instruments: 'MUSE'
    
   '''
     
   if inst == 'MUSE':
      if get_ff:
         this_fn_path = os.path.dirname(__file__)
         ref_fn = 'MUSE_specs/MUSE_spectral_resolution.txt'
         R_lam = np.loadtxt(os.path.join(this_fn_path,ref_fn), skiprows = 1)
                           
         # Fit a polynomial to this. Deg 3 works well.
         z = np.polyfit(R_lam[:,0]*10.,R_lam[:,1],3)
         p = np.poly1d(z)
          
         if show_plot:
            plt.close(99)
            plt.figure(99)
             
            lams = np.arange(4500,10000.,1)
             
            plt.plot(lams,p(lams), 'b-')            
            plt.plot(R_lam[:,0]*10.,R_lam[:,1], 'k.')
             
            plt.show()
          
      else:
         #Fit on 02.2016:
         p = np.poly1d([ -8.27037043e-09, 1.40175196e-04, -2.83940026e-01, 7.13549344e+02])
           
      return p    
                    
   else:
      raise Exception('Unknown instrument...') 
 
# ---------------------------------------------------------------------------------------- 


# -*- coding: utf-8 -*-
'''
brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
Copyright (C) 2018-2019,  F.P.A. Vogt

-----------------------------------------------------------------------------------------

This file contains several function and tools used by the brutifus routines to correct
emission line fluxes for galactic and extragalactic reddening.

Created November 2018, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au
'''
# ----------------------------------------------------------------------------------------

import numpy as np
import scipy as sp
import scipy.interpolate as interpolate
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from .brutifus_metadata import *

# ----------------------------------------------------------------------------------------      
def D_fm90(x,gamma,x0):
   ''' The UV points from Fitzpatrick & Massa (1990). 
    
   :param x: Wavelength in 1/microns.
   :type x: float
   :param gamma: The gamma factor.
   :type gamma: float
   :param x0: The x0 factor.
   :type xo: float

   '''
   
   return x**2/((x**2-x0**2)**2+x**2*gamma**2)

# ----------------------------------------------------------------------------------------      
def F_fm90(x):
   ''' The UV points from Fitzpatrick & Massa (1990). 
        
   :param x: Wavelength in 1/microns.
   :type x: float
    
   '''
   out = np.zeros_like(x)
   for (i,item) in enumerate(x):
      if item >= 5.9:    
         out[i] = 0.5392*(item-5.9)**2 + 0.05644*(item-5.9)**3
      else :
         out[i] = 0.0
   return out
   
# ----------------------------------------------------------------------------------------           
def f99_alebv(lams,rv=3.1): # lams in Angstroem
   ''' The spline-interpolated Fitzpatrick, PASP (1999) Alambda/E(B-V) function.
    
   :param lams: The wavelengths in Angstroem.
   :type lams: numpy array
   :param rv: The Rv value.
   :type rv: float
    
   '''
    
   my_lams = 1.0e4/lams 
     
   # Make sure this is an array, or the code will crash
   try :
      len(my_lams)
   except :
      my_lams = np.array([my_lams])
       
   # First, the anchor points (in 1/lambda)
   uv_anchors = np.array([3.704,3.846])
   optical_anchors = np.array([1.667, 1.828, 2.141, 2.433])
   ir_anchors = np.array([0.0,0.377, 0.820])    
             
   # ---- The parameters for the UV points ----
   c2 = -0.824 + 4.717/rv
   c1 = 2.030 - 3.007 * c2
   x0 = 4.596
   gamma = 0.99
   c3 = 3.23
   c4 = 0.41
   D = D_fm90(uv_anchors,gamma,x0)
   F = F_fm90(uv_anchors)    
   kx_V = c1 + c2 * uv_anchors + c3 * D + c4 * F
   uv_spline = kx_V + rv
    
   # ---- For the optical anchors ----
   optical_spline = np.zeros_like(optical_anchors)
   optical_spline[0] = -0.426 + 1.0044 * rv
   optical_spline[1] = -0.050 + 1.0016 * rv
   optical_spline[2] =  0.701 + 1.0016 * rv 
   optical_spline[3] = +1.208 + 1.0032 * rv - 0.00033 * rv**2 
   # WARNING : typo in F99 ?
   # See their Table 4 ... with '-1.208', it doesn't work !

   # ---- For the IR anchors ----
   ir_spline = rv/3.1 * np.array([0.0,0.265,0.829])
    
   # Ok, now, perform the spline interpolation
   wave = np.append(np.append(ir_anchors, optical_anchors), uv_anchors)
   alebv = np.append(np.append(ir_spline, optical_spline), uv_spline)   
   interp_func = interpolate.interp1d(wave, alebv, kind='cubic', 
                                       bounds_error = False) 
   # bounds_error = False -> Will return a NaN is outside the allowed range 
    
   # Finally, compute the value at the wavelengths of interest
   out = np.zeros_like(my_lams)
    
   # If lam > 2700, use the spline function
   out[lams>2700] = interp_func(my_lams[lams>2700])
   # If lam <= 2700, use the fm90 function
   Dl = D_fm90(my_lams[lams<=2700], gamma, x0)
   Fl = F_fm90(my_lams[lams<=2700])
   out[lams<=2700] = rv + c1 + c2 * my_lams[lams<=2700] + c3 * Dl + c4 * Fl
    
   return out  
   
# ----------------------------------------------------------------------------------------  
# Get E_lambda-V/E_B-V following Fischera & Dopita05. Unlike Vogt et al. (2013), follow 
# the paper more closely, and allow for varying values of Rv using polynomials given in 
# the text. In other words, use BOTH rv and rva to characterize the attenuation curve.
def fd05_elvebv(lams, rv = 3.08, rva = 4.3): # lams in Angstroem
   ''' Compute E(lambda-V)/E(B-V) according to Fischera&Dopita (2005), given Rv and Rva.
    
   This function reproduces Table 1 in the article. Unlike in Appendix 1 of Vogt et al.
   (2013), this function relies on the polynomials defined in the text to reproduce the 
   values of E(lambda-V)/E(B-V) for any Rv and Rva.
    
   :param lams: The wavelengths at which to evaluate the attenutation.
   :type lams: numpy array
   :param rv: The Rv value, following Fischera & Dopita.
   :type rv: float
   :param rva: The Rva value, following Fischera & Dopita.
   :type rva: float
    
   :return: The value of E(lambda-V)/E(B-V), as in Table 1 and Fig.3 of the article.
   :rtype: numpy array
             
   .. note:: By default, Rv = 3.08 and Rva = 4.3, which gives a curve vey close to that of 
             Calzetti (2000) below the 2000 Ansgtroem bump. Thanks to R. Sutherland at 
             ANU/RSAA for sharing his reddening code (included in MAPPINGS V), which 
             helped me implement this function properly.
             
   '''
    
   # Make sure this is an array (for the overall consistency)
   try :
      len(lams)
   except :
      lams = np.array([lams])
        
   # Rv < 3.08 is probably a bad thing ...    
   if rv < 3.08:
      sys.exit(" Rv <3.08 not supported !")
    
   # The wavelength nodes from Table 1 in microns
   lam_nodes = np.array([0.0912, 0.1053, 0.1216, 0.1550, 0.1906, 0.2175, 0.2480, 0.3000,
                          0.3650, 0.4400, 0.5480, 0.7200, 1.0300, 1.2390, 1.6490, 2.1920,
                          3.5920, 4.7770])
    
   # Do it all for the Av=1 curves. According to the paper:
   #  If the approximation obtained for AV= 1 is used for a turbulent medium where 
   # AV= 10 the accuracy is better than ∼5%.
    
   # The curve for Rv=Rva=3.08
   elv_ebv308a1 = np.array([12.066, 9.694, 7.328, 5.018, 5.214, 6.937, 4.573, 2.967, 
                             1.924, 1.000, 0.000,-0.991,-1.871,-2.171,-2.498,-2.716,
                             -2.932, -2.992])
                              
   # The x from Eqt. 8. Al/Av = (Elv/Ebv*1/Rv) + 1
   xs = elv_ebv308a1/3.08 + 1.0
    
   # The a coefficients, as polynomial using the b coefficients from Table 2.
   fa1 = np.poly1d([0.0, 1.64220, -3.62163, 3.42939, 0.00495676])
   fa2 = np.poly1d([-2.199, 5.11725, -0.668874, -0.708645, 0.0393932])
   fa3 = np.poly1d([21.5465, -19.2137, 5.17489, -0.484444, 0.0174931])
    
   # By definition, following Par.2 in Sec 3.1.:
   alphav = 3.08 
    
   # Eqt. 11
   m = 1 + 0.275 * (rv - alphav)
    
   # Eqt. 10
   alphava = m**-1*(rva - rv) + alphav
    
   # Eqt. 9
   z = (alphava-1.0)**-1
   a1 = fa1(z)
   a2 = fa2(z)
   a3 = fa3(z)
    
   # Eqt. 8
   alav = a1 * np.log10(xs)
   alav += a2 * (np.log10(xs))**2
   alav += a3 * (np.log10(xs))**3
   alav = 10.**alav
    
   # Reproduce the values of table 1. These are approximated using the polynomials, and 
   # according to the article (for the table, set rv=3.08 and rva =...):
   # If applied to situations where the absolute attenuation AV is either 1 or 10, the 
   # approximations are better than 1.3%.  In general the accuracy is somewhat lower.  
   # If the approximation obtained for AV= 1 is used for a turbulent medium where AV= 10 
   # the accuracy is better than ∼5%. The largest errors occurs at low optical depths. 
   # For x≥1 the approximation is better than ∼2%.
   elbebv = (alav-1.) * rva
    
   # Very well, now, I need to interpolate between these nodes, to get a smooth function 
   # at "any" wavelength. Use Akima splines. Fit the curve in 1/lam space, because the
   # points are a lot smoother then. Reverse their order, because x must be strictly 
   # ascending. To be clear, this fits elb_ebv as a function of 1/lambda in 1/microns
   akima_spline = sp.interpolate.Akima1DInterpolator(1./lam_nodes[::-1],elbebv[::-1]) 
    
   return akima_spline(1./(lams*1e-4))
    
# ----------------------------------------------------------------------------------------  
def cal00_ke(lams,rv):
   ''' The Calzetti (2000) law. lams in Angstroem. 
    
   .. note:: This does NOT include the 0.44 correction factor to apply for stellar 
             extinction.
             
   '''
    
   # Make sure lams is an array (for the overall consistency)
   try :
      len(lams)
   except :
      lams = np.array([lams])
    
   my_lams = 1.0e4/lams             
   ke = np.zeros_like(lams,dtype='d')
   for i in range(len(lams)):
      if lams[i]/1.0e4 < 0.63 and lams[i]/1.0e4 >= 0.12 :
         ke[i] = 2.659 * (-2.156 +1.509 * my_lams[i] - 0.198*my_lams[i]**2 +
                             0.011 * my_lams[i]**3) + rv
      elif lams[i]/1.0e4 < 2.20 and lams[i]/1.0e4 >= 0.63 :
         ke[i] = 2.659 * (-1.857 +1.040 *my_lams[i]) + rv
         
   return ke 
    
# ---------------------------------------------------------------------------------------- 
# Following Cardelli, Clayton & Mathis (1989)
def ccm89_alav(lams, rv): # lams in Angstroem
   ''' The law from Caredelli, Clayton and Mathis (1989). lams in Angstroem.
    
   '''
    
   # Make sure lams is an array 
   try :
      len(lams)
   except :
      lams = np.array([lams])

   my_lams = 1.0e4/lams
    
   a = np.zeros_like(lams, dtype ='d')
   b = np.zeros_like(lams, dtype ='d')
   alav = np.zeros_like(lams, dtype ='d')
    
   for i in range(len(my_lams)):
      x = my_lams[i]
      if 0.3 <= x and x < 1.1 :
         a[i] =  0.574 * x**1.61
         b[i] = -0.527 * x**1.61
      elif 1.1 <= x and x < 3.3 :
         y = x-1.82
         a[i] = 1. + 0.17699 * y -0.50447 * y**2 - 0.02427 * y**3 +\
                0.72085 * y**4 + 0.01979 * y**5 - 0.77530 * y**6 + 0.32999 * y**7
         b[i] = 1.41338 * y + 2.28305 * y**2 + 1.07233 * y**3 - 5.38434 * y**4 -\
                0.62251 * y**5 + 5.30260 * y**6 - 2.09002 * y**7
      elif 3.3 <= x and x<= 8 :
         if x < 5.9 : 
            fa = 0
            fb = 0
         else :
            fa = -0.04473 * (x-5.9)**2 - 0.009779 * (x-5.9)**3
            fb =  0.2130 * (x-5.9)**2 + 0.1207 * (x-5.9)**3
         a[i] =  1.752 - 0.316 * x - 0.104/((x-4.67)**2 + 0.341) + fa
         b[i] = -3.090 + 1.825 * x + 1.206/((x-4.62)**2 + 0.263) + fb
    
   alav = a + b/rv

   return alav
    
# ---------------------------------------------------------------------------------------- 
def alam(lams, ab, av, curve='f99', rv=None, rva=4.3):
   ''' Calculate the value of A_lambda for different extinction/attenuation laws.
    
   :param lams: The values at which to evaluate the attenuation/extinction.
   :type lams: numpy array
   :param ab: The extinction in B (e.g. according to NED)
   :type ab: float
   :param av: The extinction in V (e.g. according to NED) 
   :type av:
   :param curve: Which extinction/attenuation curve to use.
   :type curve: str
   :param rv: The value of Rv.
   :type rv: float
   :param rva: The value of Rva, if using curve='fd05'.
   :type rva: float
        
   .. note:: To reproduce the values given in NED, set curve='f99', and rv = 3.1. The 
             default value of Rv will depend on the curve used (if not specified). rv=3.1 
             for 'f99', rv=3.08 for 'fd05', rv=4.05 for 'cal00' and 'cal00*', and rv=3.1 
             for 'ccm89'. 'cal00*' includes a 0.44 correction factor to derive the stellar 
             extinction.  
        
   '''
   if curve == 'f99':
      if rv is None:
         rv = 3.1
      out = f99_alebv(lams,rv) * (ab-av) 
      return out
        
   elif curve == 'fd05':
      # From Fischera& Dopita (2005), see also Vogt et al. (2013).
      if rv is None:
         rv = 3.08   
      red = fd05_elvebv(lams, rv=rv, rva=rva)
      return (red + rva) * (ab-av)
        
   elif 'cal00' in curve:
      if rv is None :
         rv = 4.05
      corr = 1.0
      if curve == 'cal00*':
         corr = 0.44
      return cal00_ke(lams,rv)*(ab-av)*corr
        
   elif curve =='ccm89':
      if rv is None:
         rv = 3.1
      return ccm89_alav(lams,rv)*av      
        
# ---------------------------------------------------------------------------------------- 
def galactic_red(lams, ab, av, curve='f99', rv = None, rva = 4.5):
   ''' Calculate the galactic extinction for a given sight-line, using Ab and Av.
    
   :param lams: The values at which to evaluate the attenuation/extinction.
   :type lams: numpy array
   :param ab: The extinction in B (e.g. according to NED).
   :type ab: float
   :param av: The extinction in V (e.g. according to NED).
   :type av: float
   :param curve: Which extinction/attenuation curve to use.
   :type curve: str
   :param rv: Rv value.
   :type rv: float
   :param rva: The value of Rva, if using curve='fd05'.
   :type rva: float
         
   :return: The ratio of Flux_(unreddened)/ Flux_(observed).
   :rtype: numpy array
    
   .. note:: To follow NED, set curve='f99', and rv = 3.1. The default value of Rv will 
             depend on the curve used (if not specified). rv=3.1 for 'f99', rv=3.08 for 
             ``fd05``, rv=4.05 for 'cal00' and 'cal00*', and rv=3.1 for 'ccm89'. 
             ``cal00*`` includes a 0.44 correction factor to derive the stellar extinction.
        
   '''
   tau = alam(lams,ab,av,curve=curve,rv=rv, rva=rva) * (2.5*np.log10(np.e))**-1
          
   return np.exp(tau)
    
# ---------------------------------------------------------------------------------------- 
def extragalactic_red(lam, hahb, hahb_0, curve = 'fd05', rv = None, rva=4.3):
   ''' Calculate the extragalactic attenuation (based on Halpha/Hbeta).
   
   :param lam: The wavelength at which to evaluate the attenuation/extinction.
   :type lam: float
   :param hahb: The observed Halpha/Hbeta flux ratio. Can be a 2-D map.
   :type hahb: numpy array
   :param hahb_0: The reference theoretical unreddened Halpha/Hbeta flux ratio.
   :type hahb_0: float
   :param curve: Which extinction/attenuation curve to use.
   :type curve: str
   :param rv: Rv value.
   :type rv: float
   :param rva: The value of Rva, if using curve='fd05'. 
   :type rva: float
  
   :return: The ratio of Flux_(unattenuated)/ Flux_(attenuated)
   :rtype: numpy array
   
   .. note:: See also Vogt+ (2013), Appendix A. By default, Rv = 3.08 and Rva = 4.3, which 
             gives a curve vey close to that of Calzetti (2000) below the 2000 Ansgtroem 
             bump. Supported curves are ``fd05``, ``f99``, ``ccm89``, ``cal00`` and ``cal00*``.
             ``cal00*`` includes a 0.44 correction factor to derive the stellar extinction.
        
   '''
   if curve == 'fd05' :
      elvebv = fd05_elvebv(lam)
      ehavebv = fd05_elvebv(ha)	
      ehbvebv = fd05_elvebv(hb) 
      out = (hahb/hahb_0)**(-(elvebv + rva)/(ehavebv-ehbvebv))  
      return out
   elif curve == 'ccm89':
      if rv is None:
         rv = 3.1
      alav = ccm89_alav(lam, rv)
      aHaav = ccm89_alav(ha, rv)
      aHbav = ccm89_alav(hb, rv)
      return (hahb/hahb_0)**(-alav/(aHaav-aHbav))
        
   elif 'cal00' in curve:
      if rv == None:
         rv=4.05
      ke = cal00_ke(lam,rv)
      keHa = cal00_ke(ha,rv)
      keHb = cal00_ke(hb,rv)
      if curve == 'cal00*':
         ke *=0.44
         keHa *= 0.44
         keHb *= 0.44
      return (hahb/hahb_0)**(-ke/(keHa-keHb))
        
   elif curve == 'f99':
      if rv == None:
         rv=3.1
      alebv = f99_alebv(lam,rv)
      aHaebv = f99_alebv(ha,rv)
      aHbebv = f99_alebv(hb,rv)
      return (hahb/hahb_0)**(-alebv/(aHaebv-aHbebv))
      
# ----------------------------------------------------------------------------------------         
def hahb_to_av(hahb,hahb_0, curve = 'fd05', rv = None, rva=4.3):
   ''' Get the reddening in Av mag, from Ha/Hb ratio.
    
   :param ha_hb: The flux ratio of Halpha to Hbeta (NOT the log!)
   :type ha_hb: numpy array
   :param ha_hb_0: The theoretical value of Halpha/Hbeta. (2.86 for SB, more for AGNs)
   :type ha_hb_0: float
   :param curve: Which reddeing law to use ?
   :type curve: str
   :param rv: The value of Rv to use. None for the default.
   :type rv: float
    
   :return: The corrseponding values of Av.
   :rtype: 1D numpy array
             
   .. note:: See also Vogt+ (2013), Appendix A. By default, Rv = 3.08 and Rva = 4.3, 
             which gives a curve vey close to that of Calzetti (2000) below the 2000 
             Ansgtroem bump.
        
   '''
   if curve == 'fd05' :
      ehavebv = fd05_elvebv(ha)	
      ehbvebv = fd05_elvebv(hb) 
      out = -2.5*np.log10(hahb/hahb_0)*(rva/(ehavebv-ehbvebv))  
      return out
   
   elif curve == 'ccm89':
      if rv == None:
         rv = 3.1
      aHaav = ccm89_alav(ha, rv)
      aHbav = ccm89_alav(hb, rv)
      return -2.5*np.log10(hahb/hahb_0)*(1./(aHaav-aHbav))
        
   elif curve == 'cal00':
      if rv == None:
         rv=4.05
      keHa = cal00_ke(ha,rv)
      keHb = cal00_ke(hb,rv)
      return -2.5*np.log10(hahb/hahb_0)*(rv/(keHa-keHb))
        
   elif curve == 'f99':
      if rv == None:
         rv=3.1
      aHaebv = f99_alebv(ha,rv)
      aHbebv = f99_alebv(hb,rv)
      return -2.5*np.log10(hahb/hahb_0)*(rv/(aHaebv-aHbebv))

# ---------------------------------------------------------------------------------------- 
def check():
   ''' Check the reddening code with some plots.
   
   Make some quick plots to test the reddening functions of brutifus, by comparing with
   the original papers. And make sure I did not mess things up ...
   
   '''
   
   # 1) Check Cardelli89 and Fitzpatrick99: reproduce Fig. 7 of Fitzpatrick99    
   x = np.arange(0.0001,8.75,0.001)
   lams = 1/x*1.e4
     
   plt.close(97)
   plt.figure(97, figsize = (15,7))
   gs = gridspec.GridSpec(1,1, height_ratios=[1], width_ratios=[1])
   gs.update(left=0.1,right=0.95,bottom=0.15,top=0.9,wspace=0.05,hspace=0.05 )
   
   ax1 = plt.subplot(gs[0,0])
   
   ls = ['r','m','g','b']
   
   for (i,r) in enumerate([2.3,3.1,4.0,5.5]):
      f99_curve = f99_alebv(lams,r) - r
      ccm89_curve = (ccm89_alav(lams,r)*r) -r
      
      ax1.plot(x,f99_curve,ls[i]+'-', label=r'R$_V$ = '+np.str(r), linewidth=2)
      ax1.plot(x,ccm89_curve,'k--')

   ax1.grid(True)
   ax1.set_xlabel(r'1/$\lambda$ [$\mu$m$^{-1}$]')
   ax1.set_ylabel(r'E($\lambda$-V)/E(B-V)')
   ax1.legend(loc='lower right',fontsize=15)
   ax1.text(0.1,11,'Fitzpatrick (1999) - Fig. 7; F99 (-) vs CCM89 (- -)')
   ax1.set_xlim([0,10])
   ax1.set_ylim([-6,13])
   # Add the MUSE range because I can
   ax1.axvline(x=1./(4750./1e4),ymax=0.65,c='k',ls='-', lw=2)
   ax1.axvline(x=1./(9350./1e4),ymax=0.65,c='k',ls='-', lw=2)
   ax1.text(1.7,7,r'MUSE range (4750\AA $\rightarrow$ 9350\AA)',ha='center')
   
   plt.show()
   
   # --------------------------------------------------------------------------
   # 2) Compare with Fischera & Dopita 2005  
   
   x = np.arange(0.0001,7.5,0.1)
   lams = 1/x*1.e4
     
   plt.close(98)
   plt.figure(98, figsize=(10,8))
   gs = gridspec.GridSpec(2,1, height_ratios=[1,0.5], width_ratios=[1])
   gs.update(left=0.15,right=0.9,bottom=0.1,top=0.95,wspace=0.05,hspace=0.05 )
   
   r = 3.1
   f99_curve = f99_alebv(lams,r) - r
   ccm89_curve = (ccm89_alav(lams,r)*r) -r
   cal00_curve = cal00_ke(lams,4.05) - 4.05   
   fd05_curve = fd05_elvebv(lams,rv=3.08,rva=4.3)
   
   ax1 = plt.subplot(gs[0,0])
   ax1.plot(x,f99_curve,'r-', linewidth=1,label=r'F99, R$_V$='+np.str(r))
   ax1.plot(x,ccm89_curve,'g-',label=r'CCM89, R$_V$='+np.str(r))
   ax1.plot(x,cal00_curve,'b-',label=r'Cal00, R$_V$=4.05')
   ax1.plot(x,fd05_curve,'k--',linewidth=2, label='FD05,R$_V$=3.08, R$_V^{A}$=4.3')
   ax1.grid(True)
   ax1.set_xticklabels([])
   ax1.set_ylabel(r'E($\lambda$-V)/E(B-V)')
   ax1.legend(loc='lower right',fontsize=15)
   ax1.set_xlim([0.5,4.5])
   ax1.set_ylim([-5,5])
   ax1.axvline(x=1./(4750./1e4),ymax=0.65,c='k',ls='-', lw=2)
   ax1.axvline(x=1./(9350./1e4),ymax=0.65,c='k',ls='-', lw=2)
   ax1.text(1.7,2.5,r'MUSE range (4750\AA $\rightarrow$ 9350\AA)',ha='center')
   
   ax2 = plt.subplot(gs[1,0])
   ax2.axhline(y=0,color =(0.4,0.4,0.4), linewidth=2)
   ax2.plot(x,f99_curve-fd05_curve,'r-', linewidth=1,label=r'F99, R$_V$='+np.str(r))
   ax2.plot(x,ccm89_curve-fd05_curve,'g-',label=r'CCM89, R$_V$='+np.str(r))
   ax2.plot(x,cal00_curve-fd05_curve,'b-',label=r'Cal00, R$_V$='+np.str(r))    
   ax2.grid(True)
   ax2.set_ylabel(r'$\Delta$E($\lambda$-V)/E(B-V)')
   #ax2.legend(loc='lower right')
   ax2.set_xlim([0.5,4.5])
   #ax1.ylim([-5,5])
   ax2.set_xlabel(r'1/$\lambda$ [$\mu$m$^{-1}$]')
   ax2.axvline(x=1./(4750./1e4),c='k',ls='-', lw=2)
   ax2.axvline(x=1./(9350./1e4),c='k',ls='-', lw=2)
      
   plt.show()
   
   # --------------------------------------------------------------------------
   # 3) Reproduce Fig.1 of Calzetti (2001) 
   
   x = np.arange(0.0001,7.5,0.01)
   lams = 1/x*1.e4
   
   plt.close(99)
   plt.figure(99, figsize=(10,8))
   gs = gridspec.GridSpec(1,1, height_ratios=[1], width_ratios=[1])
   gs.update(left=0.15,right=0.9,bottom=0.1,top=0.95,wspace=0.05,hspace=0.05 )
    
   f99_2 = f99_alebv(lams,2.0)
   f99_31 = f99_alebv(lams,3.1)
   f99_5 = f99_alebv(lams,5.0)
   ccm89_2 = ccm89_alav(lams,2.0)*2.0
   ccm89_31 = ccm89_alav(lams,3.1)*3.1
   ccm89_5 = ccm89_alav(lams,5.0)*5.0
   #ccm89_curve = (ccm89_alav(lams,r)*r) -r
   cal00_curve = cal00_ke(lams,4.05)   
   fd05_curve = fd05_elvebv(lams,rv=3.08,rva=4.3)+4.3
   
   ax1 = plt.subplot(gs[0,0])
   ax1.plot(np.log10(lams/1.e4),f99_2,'g:', linewidth=1,label=r'F99, R$_V$=2.0')
   ax1.plot(np.log10(lams/1.e4),f99_31,'g-', linewidth=1,label=r'F99, R$_V$=3.1')
   ax1.plot(np.log10(lams/1.e4),f99_5,'g--', linewidth=1,label=r'F99, R$_V$=5.0')
   ax1.plot(np.log10(lams/1.e4),ccm89_2,'r:', linewidth=1,label=r'CCM89, R$_V$=2.0')
   ax1.plot(np.log10(lams/1.e4),ccm89_31,'r-', linewidth=1,label=r'CCM89, R$_V$=3.1')
   ax1.plot(np.log10(lams/1.e4),ccm89_5,'r--', linewidth=1,label=r'CCM89, R$_V$=5.0')
   ax1.plot(np.log10(lams/1.e4),cal00_curve,'b-',label=r'Cal00, R$_V$=4.05')
   ax1.plot(np.log10(lams/1.e4),fd05_curve,'k--',linewidth=2, label='FD05,R$_V$=3.08, R$_{V}^{A}$=4.3')

   ax1.grid(True)
   ax1.set_ylabel(r'A$_\lambda$/E(B-V)')
   ax1.set_xlabel(r'log[$\lambda$($\mu$m)]', labelpad=10)
   ax1.legend(loc='upper right', fontsize=15)
   ax1.set_xlim([-1,0.3])
   ax1.set_ylim([0,15])
   
   # Add the MUSE range because I can
   ax1.axvline(x=np.log10(4750./1e4),ymax=0.5,c='k',ls='-', lw=2)
   ax1.axvline(x=np.log10(9350./1e4),ymax=0.5,c='k',ls='-', lw=2)
   ax1.text(-0.15,8.1,r'MUSE range (4750\AA $\rightarrow$ 9350\AA)',ha='center')
   
       
   plt.show()
# ----------------------------------------------------------------------------------------





# -*- coding: utf-8 -*-
# pylint: disable=C0103,W0311

'''
 brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
 Copyright (C) 2018-2019,  F.P.A. Vogt

----------------------------------------------------------------------------------------------------

 This file contains WCS-related function for the higher-leve brutifus routines.

 Created August 2019, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au
'''
# --------------------------------------------------------------------------------------------------

import warnings
import os
import numpy as np
from scipy import signal

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

from astropy.io import fits
import astropy.units as u
from astropy.coordinates.sky_coordinate import SkyCoord
from astropy.stats import sigma_clipped_stats
from astropy.time import Time
from astropy.wcs import WCS

from astroquery.gaia import Gaia

from photutils import DAOStarFinder

from . import brutifus_metadata as bifus_m
from . import brutifus_plots as bifus_p

# --------------------------------------------------------------------------------------------------
def get_linear_WCS_corr(fn, obstime=None, verbose=True, suffix=None, target=None):
   ''' Identify the linear offset between Gaia and the WCS solution of a given 2D image.

   :param fn: 2D FITS image filename (and location)
   :type fn: string
   :param obstime: Observing time and date
   :type obstime: astropy.Time()
   :param suffix: the tag of this step (for plot filename)
   :type suffix: str
   :param target: target name (for plot filename)
   :type target: str
   :param verbose: print some info to screen
   :type verbose: bool

   :return: (dx,dy), the linear offsets in pixels
   :rtype: tuple

   .. note:: Assumes the image is in the primary extension.

   '''

   # Get the data
   hdu = fits.open(fn)
   header = hdu[0].header
   wl_im = hdu[0].data
   hdu.close()

   # First, let's get some of the statistics on this image
   wl_mean, wl_median, wl_std = sigma_clipped_stats(wl_im, sigma=3.0)

   if verbose:
      print('   Looking for stars in the field ...')

   # Source: https://photutils.readthedocs.io/en/stable/detection.html
   # Get ready to look for stars in there
   daofind = DAOStarFinder(fwhm=3.0, threshold=5. * wl_std)

   # Start the search
   sources = daofind(wl_im - wl_median)

   # Create a list of sources in pixels
   sources_list = np.array([sources['xcentroid'], sources['ycentroid']])

   # Reshape this properly as Nx2
   sources_list = sources_list.T

   # Store the 2D WCS info for later
   wcs2d = WCS(header)

   # Find the coordinate at the center of the image.
   # To be robust (E.g. OCam), should be CRVAl 1/2 + CDELT1/2*(0.5*NAXIS 1/2 - CRPIX1/2)
   # Fixed v. 2019.08.1
   # Credits to J. Suherli for stumbling upon and reporting the bug.
   core_coord = SkyCoord(ra=header['CRVAL1'] +
                            header['CD1_1'] * (0.5*header['NAXIS1'] - header['CRPIX1']),
                         dec=header['CRVAL2'] +
                             header['CD2_2'] * (0.5*header['NAXIS2'] - header['CRPIX2']),
                         unit=(u.deg, u.deg), frame='icrs')
   search_radius = 0.75 * max([np.abs(header['NAXIS1'] * header['CD1_1']*u.deg),
                               np.abs(header['NAXIS2'] * header['CD2_2']*u.deg)])

   if verbose:
      print('   Querying GAIA ...')

   # Make it a sync search, because I don't think I need more than 2000 targets ...
   j = Gaia.cone_search_async(core_coord, search_radius, verbose=False)
   r = j.get_results()

   if len(r) >= 2000:
      warnings.warn('   Maximum number of GAIA entries returned. Some stars might be missing.')

   # Some of these proper motions are NaN's ... need to deal with this before feeding them
   # to SkyCoord
   r['pmra'].fill_value = 0
   r['pmdec'].fill_value = 0

   # Turn this into a "catalogue" of SkyCoord
   gaia_cat = SkyCoord(ra=r['ra'], dec=r['dec'], unit=(u.deg, u.deg), frame='icrs',
                       pm_ra_cosdec=r['pmra'].filled(),
                       pm_dec=r['pmdec'].filled(),
                       obstime=[Time(item, format='decimalyear') for item in r['ref_epoch']],
                       equinox='J2000',
                       distance=1.e3 * u.pc, # Assumes the default distance
                       )

   # Propagate Gaia to the observation date!
   if not obstime is None:
      if verbose:
         print('   Propagating proper motions to %s ...' % (obstime))
      gaia_cat = gaia_cat.apply_space_motion(new_obstime=obstime)

   # Clean this up, to only show the points within the image
   gaia_cat = gaia_cat[wcs2d.footprint_contains(gaia_cat)]

   # Transform this in pixel coords
   gaia_cat_pix = wcs2d.wcs_world2pix(gaia_cat.ra, gaia_cat.dec, 0)

   # Reshape it properly
   gaia_cat_pix = np.array(gaia_cat_pix).T

   # Ok, let's make two fake image from the source list, that I then cross-correlate
   nx = header['NAXIS1']
   ny = header['NAXIS2']

   im_ref = np.zeros((ny, nx))
   im_obs = np.zeros((ny, nx))

   # Fill the two images with fake stars
   # v2019.08.2: this gets slow with large numbers of stars
   # (thanks Janette Suherli for reporting this)
   # let's be smarter about this ... only create stars on small sub-arrays,
   # and add these to the final image (rather than creating full images each time).
   # Note: the FWHM of the stars in the fake images is fixed in pixels ... should I change this ?

   fwhm_pix = 2

   # Let's get ready to fill the images.

   for (im, entries) in [(im_obs, sources_list), (im_ref, gaia_cat_pix)]:

      # Let's create the fake observed image ...
      for i in range(len(entries)):

         # Pixel ID closest from the star centroid
         xc = np.floor(entries[i][0])
         yc = np.floor(entries[i][1])

         # If we are too close from the edge, skip the star
         # (otherwise I have array size issues)
         # That's the easy way out of course ... !
         if (xc < 2 * fwhm_pix + 3) or (xc > nx - (2 * fwhm_pix + 3)) or \
            (yc < 2 * fwhm_pix + 3) or (yc > ny - (2 * fwhm_pix + 3)):
            continue

         # Find the x and y indices of the sub-array pixels
         x_ind, y_ind = np.meshgrid(np.arange(xc-(2 * fwhm_pix + 1), xc+(2 * fwhm_pix + 1), 1),
                                    np.arange(yc-(2 * fwhm_pix + 1), yc+(2 * fwhm_pix + 1), 1))

      # For each of these, compute the distance to the star center
      d = np.sqrt((x_ind - entries[i][0])**2 + (y_ind - entries[i][1])**2)

      # And now add the star to the full image
      im[y_ind.astype(int), x_ind.astype(int)] += np.exp(-((d)**2 / (2.0 * fwhm_pix**2)))


   # Having done that, let's convolve the two images
   # Credit to Nick Whyborn @ ALMA for the idea
   # Don't forget to flip one of the image to get the correlation
   im_corr = signal.convolve(im_ref, im_obs[::-1, ::-1], mode='same', method='auto')

   # Now, let's find the brightest pixel
   peak_loc = np.unravel_index(im_corr.argmax(), im_corr.shape)

   # Let's also find the "fine" position of the peak with photutils
   mean, median, std = sigma_clipped_stats(im_corr, sigma=10.0)
   daofind = DAOStarFinder(fwhm=5.0, threshold=10.*std)
   peaks = daofind(im_corr)

   # Add a clean error, in case I find no peak.
   if not peaks: # Neat trick ... empty sequences == False !
      raise Exception('Ouch! No WCS correlation peak identified.')

   # Ok, which one is the brightest peak ?
   peak_id = np.argmax(peaks['peak'])
   peak_loc_fine = (peaks['xcentroid'][peak_id], peaks['ycentroid'][peak_id])

   # Issue a warning if the brightest peak is too far from the brightest pixel
   if np.abs(peak_loc_fine[0] - peak_loc[1]) > 2 or \
      np.abs(peak_loc_fine[1] - peak_loc[0]) > 2:

      warnings.warn('WCS correlation peak does not coincide with the brightest pixel!')

   # Ok, so what are the offsets I should apply ?
   dx = peak_loc_fine[0]-np.floor(nx/2) # Added the floor... it is required by 'same' ?
   dy = peak_loc_fine[1]-np.floor(ny/2) # v2019.07.1

   if verbose:
      print('   Best offset: %+.2f  %+.2f pixels' % (dx, dy))

   # Very well, now apply the corrections to the image
   fits.setval(fn, 'CRPIX1', value=header['CRPIX1'] - dx, ext=0)
   fits.setval(fn, 'CRPIX2', value=header['CRPIX2'] - dy, ext=0)

   # Let's make a plot to show how good I'm doing with getting a reasonable offset
   plt.close(99)
   fig99 = plt.figure(99, figsize=(6.94, 7))

   gs = gridspec.GridSpec(1, 1, height_ratios=[1], width_ratios=[1],
                          left=0.12, right=0.98, bottom=0.12, top=0.9,
                          wspace=0.1, hspace=0.1)

   ax99 = plt.subplot(gs[0, 0])

   ax99.imshow(im_corr, cmap='gray', origin='lower', interpolation='nearest')

   ax99.plot(peak_loc_fine[0], peak_loc_fine[1], marker=bifus_p.crosshair(pa=0, inner_r=0.5),
             c='darkorange', markersize=50)

   ax99.set_aspect('equal')
   ax99.set_xlabel('x [pix]')
   ax99.set_ylabel('y [pix]')
   ax99.set_title('Correlation peak: (%+.2f,%+.2f) spaxels' % (dx, dy))

   fig99.savefig(os.path.join(bifus_m.plot_loc, suffix + '_' + target +
                              '_get-WCS-offsets.pdf'))
   plt.close(99)

   # Now plot the white-light image to show how good I did ...
   (fig, ax, ofn) = bifus_p.make_2Dplot(fn, ext=0,
                                        ofn=os.path.join(bifus_m.plot_loc, suffix + '_' + target +
                                                         '_check-WCS.pdf'),
                                        plims=[10, 99],
                                        cmap=None, cblabel=None)

   # And add them to the plot
   ax.scatter(gaia_cat.ra, gaia_cat.dec, marker='o', color='darkorange',
              s=20, facecolor='none', transform=ax.get_transform('world'))

   # Show the stars I found in the data ... these are in pix coordinates ...
   # -> always "right" irrespective of the WCS!
   ax.scatter(sources['xcentroid'], sources['ycentroid'],
              color='crimson', s=50, marker='o', facecolor='none')

   # Save the updated plot
   fig.savefig(ofn)

   return (dx, dy)

# --------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
'''
brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
Copyright (C) 2018-2019,  F.P.A. Vogt
 
-----------------------------------------------------------------------------------------
 
This file contains some convenience functions for brutifus. 

Created July 2019, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au
 
-----------------------------------------------------------------------------------------
  
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
'''

from . import brutifus_metadata as bifus_m
from .brutifus_version import __version__
from . import brutifus as bifus

import os
import shutil
import argparse

# Use argparse to make brutifus user friendly ---------------------------------------------
parser = argparse.ArgumentParser(description=''' Aids in the post-processing of 3D datacubes. ''',
                                 epilog =' Full documentation: %s' % ('http://fpavogt.github.io/brutifus'),
                                 formatter_class=argparse.RawTextHelpFormatter)

parser.add_argument('-v', '--version', action='version', version=('%s'%__version__))                   
                    
parser.add_argument('-s', '--setup', action = 'store_true', 
                    help = 'Create a local copy of the parameter and execution files '+
                            'needed to run brutifus.')

parser.add_argument('-e', '--execute', action='store', nargs = 2, 
                    metavar = ('procsteps_file', 'params_file'),
                    #type=open, 
                    help='run the steps specified in "procsteps_file" '+
                         'with the high-level parameters specified in "params_file".') 

def main():
   # What did the user type in ?
   args = parser.parse_args()
   
   # Setup the scene ?
   if args.setup:
      
      print('')
      
      # Very well, let's copy the files at the current location
      for f in [bifus_m.bifus_params, bifus_m.bifus_procsteps]:
         if not os.path.isfile(os.path.join('.',f)):
          
            print ('   - creating the local copy of %s ...' % (f))
            shutil.copyfile(os.path.join(bifus_m.bifus_dir,'exec_scripts',f), f)
            
         
      # Let's also create the working directories if needed
      for d in [bifus_m.plot_loc, bifus_m.prod_loc]:
         if not os.path.isdir(os.path.join('.',d)):
            
            print ('   - creating the local directory "%s" ...' % (d))
            os.mkdir(os.path.join('.',d))
            
      # Don't forget the tmp one
      tmp_dir = os.path.join('.', bifus_m.prod_loc, bifus_m.tmp_loc)
      if not os.path.isdir(tmp_dir):
      
         print ('   - creating the local directory "%s" ...' % (tmp_dir))
         os.mkdir(tmp_dir)
       
             
      print('')
      print('brutifus setup complete.')
   
   if args.execute:
      # Start the processing
      bifus.run(args.execute[0], args.execute[1])

   
# Start of the interactive part ----------------------------------------------------------
if __name__ == "__main__":

   main()
            
            
      
      
   
   


# -*- coding: utf-8 -*-
'''
brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
Copyright (C) 2018-2019,  F.P.A. Vogt

-----------------------------------------------------------------------------------------

This file contains some global metadata used throughout the brutifus code.

Created November 2018-2019, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au
'''
# ----------------------------------------------------------------------------------------

import numpy as np
import os

import matplotlib.pyplot as plt

# ---| Some basic parameters |------------------------------------------------------------

# Where are we located ?
bifus_dir = os.path.dirname(__file__) # Get the project directory from the file location!

# Name of the parameters files
bifus_params = 'params_brutifus.yaml'
bifus_procsteps = 'procsteps_brutifus.yaml'

# Name of default storage spaces
plot_loc = 'brutifus_plots'
prod_loc = 'brutifus_products'
tmp_loc =  'brutifus_tmp'


def get_fn_list_fn(target):
   ''' Returns the filename of the storage pickle for all the filenames used by the code
      
      :param target: name of the target/object to be processed
      :type target: string 
   
   
      :return: the pickled dictionary filename
      :rtype: string
   '''
      
   return 'bifus_fn-list_%s.pkl' % (target) # Name of the dictionary for filenames

# ---| Plotting parameters |--------------------------------------------------------------

usetex = True
plotstyle = os.path.join(bifus_dir,'mpl_styles','brutifus_plots.mplstyle')

fig_width = 6.94 # In inches
margin_left = 1.3 # in inches
margin_right = 0.2 # in inches
margin_bottom = 0.75 # in inches
margin_top = 0.7 # in inches
cb_thick =  0.3 # in inches

# ---| Information about the file format for the different instruments |------------------

ffmt = {'MUSE': {'data':1, 'var':2, 'badpix':None, 
                 'funit':r'F$_\lambda$ [$10^{-20}$ erg s$^{-1}$ cm$^{-2}$ \AA$^{-1}$]'},
        'WiFeS':{'data':0, 'error':1, 'badpix':2, 
                 'funit':r'F$_\lambda$ [erg s$^{-1}$ cm$^{-2}$ \AA$^{-1}$]'},
       }


      


from setuptools import setup, find_packages  # Always prefer setuptools over distutils
import os

# Run the version file
v = open(os.path.join('.','brutifus','brutifus_version.py'))
version = [l.split("'")[1] for l in v.readlines() if '__version__' in l][0]

setup(
   name='brutifus',
   version=version,
   author='F.P.A. Vogt',
   author_email='frederic.vogt@alumni.anu.edu.au',
   packages=['brutifus',],
   url='http://fpavogt.github.io/brutifus/',
   download_url='https://github.com/fpavogt/brutifus/archive/master.zip',
   license='GNU General Public License',
   description='Python module to process IFU datacubes.',
   long_description=open('README').read(),
   python_requires='>=3',
   install_requires=[
      "numpy >= 1.14.2",
      "scipy >= 1.1.0",
      "matplotlib >= 3.0.0",
      "astropy >= 3.0",
      "astroquery >= 0.3.4",
      "statsmodels >= 0.9.0",
      "photutils >= 0.6",
   ],
   
   entry_points={'console_scripts': ['brutifus=brutifus.__main__:main']},
    
   classifiers=[
   # How mature is this project? Common values are
   #   3 - Alpha
   #   4 - Beta
   #   5 - Production/Stable
   'Development Status :: 3 - Alpha',

   # Indicate who your project is intended for
   'Intended Audience :: Science/Research',
   'Topic :: Scientific/Engineering :: Astronomy',

   # Pick your license as you wish (should match "license" above)
   'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    
   # Specify the Python versions you support here. In particular, ensure
   # that you indicate whether you support Python 2, Python 3 or both.
   'Programming Language :: Python :: 3.7',
   ],
    
   include_package_data=True, # So that non .py files make it onto pypi, and then back !
   #package_data={
   #     'example_files':['example_files/*'],
   #     'docs':['../docs/build']
   #    }
)


# -*- coding: utf-8 -*-
# disable bad variable names, 3 spaces for offset
# pylint: disable=C0103,W0311
'''
brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
Copyright (C) 2018-2019,  F.P.A. Vogt

----------------------------------------------------------------------------------------------------

This file contains the master brutifus routines. Most of these routines call sub-routines,
after setting the scene/loading datasets/etc ...

Any processing step MUST have a dediacted routine in this file called 'run_XXX', which can
then refer to any existing/new brutifus/Python module.

Created November 2018, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au

----------------------------------------------------------------------------------------------------

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
# --------------------------------------------------------------------------------------------------

import sys
import os
import warnings
from functools import partial
import datetime
import multiprocessing
import pickle

import numpy as np

#from matplotlib import MatplotlibDeprecationWarning
#warnings.filterwarnings("error",category=MatplotlibDeprecationWarning)
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

from astropy.io import fits
from astropy.time import Time
from astropy import log as astropy_log

import yaml

# Import brutifus-specific tools
from . import brutifus_tools as bifus_t
from . import brutifus_cof as bifus_cof
from . import brutifus_plots as bifus_p
from . import brutifus_red as bifus_red
from . import brutifus_wcs as bifus_wcs
from . import brutifus_metadata as bifus_m
from .brutifus_version import __version__

# Disable the annoying warnings from astropy
astropy_log.setLevel('WARNING') # Disable pesky INFO flags from aplpy

# For debugging purposes: turn all warnings into errors
#warnings.filterwarnings("error", category=AstropyDeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

# Also filter stupid GAIA warnings
warnings.filterwarnings("ignore", module='astropy.io.votable.tree')

# --------------------------------------------------------------------------------------------------
def run(procsteps_fn, params_fn):
   ''' Run a given set of brutifus processing steps with certain parameters.

       :param procsteps_fn: filename containing the processing steps
       :type procsteps_fn: string
       :param params_fn: filename containing the scientific parameters
       :type params_fn: string

   '''

   # Start keeping track of the time
   start_time = datetime.datetime.now()

   # Check that I was fed some proper parameters
   if not os.path.isfile(params_fn):
      raise Exception('Failed to load the parameter file %s.' % (params_fn))

   if not os.path.isfile(procsteps_fn):
      raise Exception('Failed to load the procsteps file %s.' % (procsteps_fn))

   # Load the parameter file
   params = yaml.load(open(params_fn), Loader=yaml.SafeLoader)

   # Load the proc steps
   procsteps = yaml.load(open(procsteps_fn), Loader=yaml.SafeLoader)

   # Disable the use of system-Latex if required ... (Why would anyone ask such a thing?!)
   if not params['systemtex']:
      bifus_m.usetex = False
      bifus_m.plotstyle = os.path.join(bifus_m.bifus_dir, 'mpl_styles',
                                         'brutifus_plots_nolatex.mplstyle')
   else:
      bifus_m.usetex = True
      bifus_m.plotstyle = os.path.join(bifus_m.bifus_dir, 'mpl_styles',
                                         'brutifus_plots.mplstyle')

   # Set the chosen plot style
   plt.style.use(bifus_m.plotstyle)

   # Is there a dictionary of filenames already in place ? If not, create one
   fn = os.path.join(bifus_m.prod_loc, bifus_m.get_fn_list_fn(params['target']))

   if not os.path.isfile(fn):
      if params['verbose']:
         print('-> This looks like a fresh run. Creating an emtpy dictionary of filenames.')
         print('-> Storing it in %s.' % (fn))

      # For now, all I know is where the raw data is located
      fn_list = {'raw_cube': os.path.join(params['data_loc'], params['data_fn']),
                }

      # Make a quick check to make sure we can find the data!
      if not os.path.isfile(fn_list['raw_cube']):
         raise Exception('Raw file not found: %s' % fn_list['raw_cube'])

      # Save it
      f = open(fn, 'wb')
      pickle.dump(fn_list, f)
      f.close()

   # Execute the recipe, by calling all the individual master step functions
   prev_suffix = None
   for step in procsteps:
      step_name = step['step']
      step_run = step['run']
      step_suffix = step['suffix']
      step_args = step['args']
      func_name = 'run_' + step_name

      # Call a function based on a string ... pretty sweet !
      #func = getattr(func_name)
      func = globals()[func_name]

      if step_run:
         # Here, I want to maintain a dictionary of filenames, to be used accross functions
         # For each step, load the dictionary, feed it to the function, and save it update
         # it when it's all done. Each function returns the updated dictionary !
         fn = os.path.join(bifus_m.prod_loc, bifus_m.get_fn_list_fn(params['target']))
         f = open(fn,'rb')
         fn_list = pickle.load(f)
         f.close()

         # Launch the function
         fn_list = func(fn_list, params, suffix=step_suffix, **step_args)

         # Save the updated dictionary of filenames
         f = open(fn,'wb')
         fn_list = pickle.dump(fn_list,f)
         f.close()

   # All done !
   duration = datetime.datetime.now() - start_time
   print(' ')
   print('All done in %.01f seconds.' % duration.total_seconds())
   print(' ')


# --------------------------------------------------------------------------------------------------
def run_adjust_WCS(fn_list, params, suffix = None, name_in = None, name_out = None):
   ''' Adjust the WCS solution of the cube using Gaia.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params:  The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: str
   :param name_out: name tag to identify which cube comes out of the routine
   :type name_out: str

   :return: The updated dictionary of filenames.
   :rtype: dict

   .. note:: Proper motions for the GAIA entries are propagated to the DATE-OBS of the data.
             Source identification in white-light image based on :class:`photutils.DAOStarFinder`

   .. todo:: Avoid using a hard-coded pixel FWHM for the cross-correlation?

   '''

   if params['verbose']:
      print('-> Adjusting the cube WCS using Gaia.')

   # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Build a white-light image
   wl_im = np.nansum(data, axis=0)

   # Save this 2D white-light image
   #hdu0 = fits.PrimaryHDU(None, header0)
   hdu1 = fits.PrimaryHDU(wl_im)
   # Make sure the WCS coordinates are included as well
   hdu1 = bifus_t.hdu_add_wcs(hdu1, header_data)
   # Also include a brief mention about which version of brutifus is being used
   hdu1 = bifus_t.hdu_add_brutifus(hdu1, suffix)

   # Write the file!
   hdu = fits.HDUList(hdus = [hdu1])

   fn_list['white_light'] = os.path.join(bifus_m.prod_loc,
                                         suffix+'_'+params['target']+'_white_light.fits')
   hdu.writeto(fn_list['white_light'], overwrite=True)


   # Get the dx dy corrections to be applied
   (dx, dy) = bifus_wcs.get_linear_WCS_corr(fn_list['white_light'],
                                            obstime=Time(header0['DATE-OBS']),
                                            verbose=params['verbose'],
                                            suffix=suffix, target=params['target'])

   # Then also correct the WCS in the datacube
   hdu0 = fits.PrimaryHDU(None, header0)
   hdu1 = fits.ImageHDU(data, header_data)
   hdu2 = fits.ImageHDU(error, header_error)

   for hdu in [hdu1,hdu2]:
      # Also include a brief mention about which version of brutifus is being used
      hdu = bifus_t.hdu_add_brutifus(hdu, suffix)

      # And update the WCS value
      hdu.header['CRPIX1'] = hdu.header['CRPIX1'] - dx
      hdu.header['CRPIX2'] = hdu.header['CRPIX2'] - dy

   # Add the filename to the dictionary of filenames
   fn_list[name_out] = os.path.join(bifus_m.prod_loc,
                                    suffix + '_' + params['target'] + '_wcs-corr.fits')
   hdu = fits.HDUList(hdus=[hdu0, hdu1, hdu2])
   hdu.writeto(fn_list[name_out], overwrite=True)

   return fn_list

# --------------------------------------------------------------------------------------------------
def run_crude_snr_maps(fn_list, params, suffix = None, name_in = None,
                       zcorr_lams = False):
   ''' Computes a crude S/N ratio for a continuum range or emission line.

   This function computes the SNR maps for the continuum or emission lines.
   It also creates a map of spaxels with any signal at all.
   The resulting maps are saved to a fits file with full header and WCS coordinates.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params:  The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param name_in: name tag to identify which cube to use to run the routine.
   :type name_in: str
   :param zcorr_lams: True -> correct the wavelength range for the target redshift.
   :type zcorr_lams: bool

   :return: The updated dictionary of filenames.
   :rtype: dict

   .. note:: This function is a "first guess" of the SNR for latter use in the code. A more
             reliable measure of the SNR for the emission line should be computed after they
             have been fitted.
   '''

   if params['verbose']:
      print('-> Computing the SNR maps.')

   # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Take redshift into account ?
   if zcorr_lams:
      lams /= params['z_target']+ 1.

   # Prepare a storage list
   snrs = []

   # Here, hide those stupid warnings about all-NaNs slices
   #with warnings.catch_warnings():
      #warnings.simplefilter("ignore", category=RuntimeWarning)

   for r in params['snr_ranges']:

         # The signal
         if r[-1] == 'c':
            s = np.nanmedian(data[(lams>=r[0])*(lams<=r[1]),:,:], axis = 0)

         elif r[-1] == 'e':
            s = np.nanmax(data[(lams>=r[0]) * (lams<=r[1]), :,:], axis = 0)

         else:
            raise Exception('S/N calculation type unknown: %s' % r[-1])

         # The noise = STD over the range -> NOT strictly correct for high S/N stars !!!
         n = np.nanstd(data[(lams>=r[0])*(lams<=r[1]),:,:], axis = 0)

         # Compute S/N
         snr = s/n

         # Ensure this is always >= 0
         snr[snr<0] = 0

         # Store it for later
         snrs += [snr]

   # And create a map with just spaxels that have any data (i.e. have been observed).
   anything = np.ones_like(data[0,:,:])
   anything[np.all(np.isnan(data),axis=0)] = np.nan
       
   # Very well, now let's create a fits file to save this as required.
   hdu0 = fits.PrimaryHDU(None,header0)
   hdus = [hdu0]
   
   # Add the regions where I have data ... always useful !
   hdu = fits.ImageHDU(anything)
   # Make sure the WCS coordinates are included as well
   hdu = bifus_t.hdu_add_wcs(hdu, header_data)
   # Also include a brief mention about which version of brutifus is being used
   hdu = bifus_t.hdu_add_brutifus(hdu, suffix)
   # For reference, also include the line/region this maps are based on
   hdu.header['B_SNRANG'] = ('%.1f-%.1f' % (lams[0],lams[-1]), 'spectral range (A) used for SNR')
   hdu.header['B_SNTYPE'] = ('x', 'binary map: NaN= no data, 1 = valid spectra')

   hdus += [hdu]

   # Now also loop through all the maps I have created
   for (i,r) in enumerate(params['snr_ranges']):

      hdu = fits.ImageHDU(snrs[i])

      # Make sure the WCS coordinates are included as well
      hdu = bifus_t.hdu_add_wcs(hdu, header_data)

      # Also include a brief mention about which version of brutifus is being used
      hdu = bifus_t.hdu_add_brutifus(hdu, suffix)

      # For reference, also include the line/region this maps are based on
      hdu.header['B_SNRANG'] = ('%.1f-%.1f' % (r[0],r[1]), 'spectral range (A) used for SNR')
      hdu.header['B_SNTYPE'] = ('%s' % r[-1], '"c"ontinuum, or "e"mission')

      hdus += [hdu]

   # Write the file!
   hdu = fits.HDUList(hdus=hdus)

   fn_list['snr_maps'] = os.path.join(bifus_m.prod_loc,
                                      suffix+'_'+params['target']+'_snr-maps.fits')

   hdu.writeto(fn_list['snr_maps'], overwrite = True)

   # Make some plots
   # First, plot the region with any signal at all
   bifus_p.make_2Dplot(fn_list['snr_maps'], 1,
                       os.path.join(bifus_m.plot_loc,
                                    suffix + '_' + params['target'] +
                                    '_valid_spectra.pdf'),
                       #contour_fn = None,
                       #contour_ext = None,
                       #contour_levels = [None],
                       vmin = 0, vmax = 1.5,
                       cmap = None, cblabel = None, scalebar = params['scalebar'])

   # Alright, let's take out the big guns ...
   for (i,r) in enumerate(params['snr_ranges']):

      bifus_p.make_2Dplot(fn_list['snr_maps'], i+2,
                          os.path.join(bifus_m.plot_loc,
                                       suffix + '_' + params['target'] +
                                       '_snr_%.1f-%.1f.pdf' % (r[0],r[1])),
                          vmin = 0, vmax = 50, cmap = 'alligator_r',
                          cblabel = r"S/N ('%s') %.1f\AA-%.1f\AA" % (r[-1],r[0],r[1]),
                          scalebar = params['scalebar'])

   return fn_list

# --------------------------------------------------------------------------------------------------
def run_plot_BW(fn_list, params, suffix=None, name_in=None,
                bands=[[7500., 9300.]],
                conts=[[None, None]],
                stretches=['arcsinh'],
                plims=[[10., 99.5]],
                vlims=[[None,None]],
                gauss_blurs=[None]
               ):
   ''' Make some B&W images from slices in the cube.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params:  The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: string
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: string
   :param bands: A list of wavelength pairs.
   :type bands: list
   :param conts: A list of wavelength pairs channels associated continuum regions.
   :type conts: list
   :param stretches: The stretches to apply to the data, e.g. 'linear', 'log', 'arcsinh'.
   :type stretches: list
   :param plims: The limiting percentiles for the plot
   :type plims: list
   :param vlims: The limiting values for the plot (superseeds stretch_plims)
   :type vlims: list
   :param gauss_blurs: Radius of gaussian blur, None for nothing
   :type gauss_blurs: list

   :return: The updated dictionary of filenames.
   :rtype: dictionary

   '''

   if params['verbose']:
      print('-> Creating %i B&W images.' % (len(bands)))

   # Doing some preliminary checks
   for item in [conts, stretches, plims, vlims, gauss_blurs]:
      if len(item) < len(bands):
         item *= int(np.ceil(len(bands)/len(item)))

   # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Step 1: Construct individual images for each band
   for (i, band) in enumerate(bands):

      # get the data out
      scidata = data[(lams >= band[0]) * (lams <= band[1]), :, :]

      # Get the continuum
      if None not in conts[i]:
         contdata = data[(lams >= conts[i][0]) * (lams <= conts[i][1]), :, :]
      else:
         contdata = 0

      # Subtract the continuum
      scidata = np.nansum(scidata, axis=0) - np.nansum(contdata, axis=0)

      fn = os.path.join(bifus_m.prod_loc, suffix + '_' + params['target'] + '_' + name_in +
                        '_BW_%i-%i.fits' % (band[0], band[1]))
      hdu = fits.PrimaryHDU(scidata)
      # Add the wcs info
      hdu = bifus_t.hdu_add_wcs(hdu, header_data)
      # Add the brutifus info
      hdu = bifus_t.hdu_add_brutifus(hdu, suffix)
      outfits = fits.HDUList([hdu])
      outfits.writeto(fn, overwrite=True)

      ofn = os.path.join(bifus_m.plot_loc, suffix + '_' + params['target'] + '_' + name_in +
                         '_BW_%i-%i.pdf' % (band[0], band[1]))

      # Great, I am now ready to call the plotting function
      bifus_p.make_2Dplot(fn, ext = 0, ofn = ofn,
                           stretch = stretches[i],
                           plims = [plims[i][0], plims[i][1]],
                           vlims = [vlims[i][0], vlims[i][1]],
                           gauss_blur = gauss_blurs[i],
                           cmap = None,
                           #title = r'\smaller %s-%s\,\AA\' % (band[0],band[1]),
                           #scalebar = params['scalebar']
                           )

   return fn_list

# --------------------------------------------------------------------------------------------------
def run_plot_RGB(fn_list, params, suffix=None, name_in=None,
                 bands=[[7500., 9300., 6000., 7500., 4800., 6000.]],
                 conts=[[None, None, None]],
                 stretches=[['arcsinh','arcsinh','arcsinh']],
                 plims=[[10., 99.5, 10., 99.5, 10., 99.5]],
                 vlims=[[None, None, None, None, None, None]],
                 gauss_blurs=[[None, None, None]],
                 ):
   ''' Make some RGB images from slices in the raw cube.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params:  The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: string
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: string
   :param bands: A list of 3 wavelength pairs.
   :type bands: list
   :param conts: A list of 3 wavelength pairs: channels associated continuum regions.
   :type conts: list
   :param stretches: The stretches to apply to each slice: 'linear', 'log', 'arcsinh', ...
   :type stretches: list
   :param plims: The limiting percentiles for the plot
   :type plims: list
   :param vlims: The limiting values for the plot (superseeds stretch_plims)
   :type vlims: list
   :param gauss_blurs: Radius of gaussian blur, None for nothing
   :type gauss_blurs: list

   :return: The updated dictionary of filenames.
   :rtype: dictionary

   '''

   # First, I need to generate 3 individual fits files for each band
   if params['verbose']:
      print('-> Creating %i RGB images.' % (len(bands)))

   # Doing some preliminary checks
   for item in [conts, stretches, plims, vlims, gauss_blurs]:
      if len(item) < len(bands):
         item *= int(np.ceil(len(bands)/len(item)))

   # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Step 1: Construct individual images for each band
   for (i,band) in enumerate(bands):

      fns = []

      for j in range(3):
         # get the data out
         scidata = data[(lams >= band[0 + 2*j]) * (lams <= band[1 + 2*j]), :, :]
         if None not in conts[i][0 + 2*j:2 + 2*j]:
            contdata = data[(lams >= conts[i][0 + 2*j]) * (lams <= conts[i][1 + 2*j]), :, :]
         else:
            contdata = 0

         scidata = np.nansum(scidata, axis=0) - np.nansum(contdata, axis=0)

         fn = os.path.join(bifus_m.prod_loc,'RGB_tmp_%i.fits' % j)
         fns.append(fn)
         hdu = fits.PrimaryHDU(scidata)
         # Add the wcs info
         hdu = bifus_t.hdu_add_wcs(hdu, header_data)
         outfits = fits.HDUList([hdu])
         outfits.writeto(fn, overwrite=True)

      ofn = os.path.join(bifus_m.plot_loc, suffix + '_' + params['target'] + '_' + name_in +
                         '_RGB_%i-%i_%i-%i_%i-%i.pdf' % (band[0], band[1], band[2], band[3],
                                                         band[4], band[5]))

      # Great, I am now ready to call the plotting function
      bifus_p.make_RGBplot(fns, ofn,
                           stretch = stretches[i],
                           plims = plims[i],
                           vlims = vlims[i],
                           gauss_blur = gauss_blurs[i],
                           title = r'\smaller R: %s-%s\,\AA\  G: %s-%s\,\AA\  B: %s-%s\,\AA' %
                                   (band[0], band[1], band[2], band[3], band[4], band[5]),
                           # scalebar = params['scalebar']
                           )

      # And remember to delete all the temporary files
      for f in fns:
         os.remove(f)

   return fn_list

# --------------------------------------------------------------------------------------------------
def run_sky_sub(fn_list, params, suffix = None, name_in = None, name_out = None):
   ''' Performs  a manual sky subtraction, when the observation strategy was flawed.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params: The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: str
   :param name_out: name tag to identify which cube comes out of the routine
   :type name_out: str

   :return: The updated dictionary of filenames.
   :rtype: dict  

   '''

   if params['verbose']:
      print('-> Subtracting the sky ...')

   # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Assemble a 2D map of the sky spaxels
   sky_spaxels = np.zeros_like(data[0,:,:])
   cys,cxs = np.mgrid[0:header_data['NAXIS2'],0:header_data['NAXIS1']]

   #  Flag all the (sky) spaxels within the user-defined aperture
   for sr in params['sky_regions']:

      if len(sr) == 3: # Circular aperture

         d = np.sqrt( (cxs-sr[0])**2 + (cys-sr[1])**2)
         sky_spaxels[d<=sr[2]] = 1

      elif len(sr) ==4: # Square aperture

         sky_spaxels[sr[1]:sr[1]+sr[3]+1, sr[0]:sr[0]+sr[2]+1] = 1

   # Very well, assemble the sky spectrum now
   sky_spec = np.array([np.nanmedian(plane[sky_spaxels==1]) for plane in data[:]])

   # Make a descent plot of the sky spectrum
   plt.close(1)
   plt.figure(1, figsize=(14.16,4))
   gs = gridspec.GridSpec(1,1, height_ratios=[1], width_ratios=[1],
                          left=0.08, right=0.98, bottom=0.18, top=0.98, wspace=0.02, hspace=0.03)

   ax1 = plt.subplot(gs[0,0])

   ax1.semilogy(lams,sky_spec, 'k-', drawstyle='steps-mid', lw=0.8)

   ax1.set_xlim((lams[0],lams[-1]))
   ax1.set_xlabel(r'Observed wavelength [\AA]')
   ax1.set_ylabel(bifus_m.ffmt[params['inst']]['funit'])

   plt.savefig(os.path.join(bifus_m.plot_loc, suffix+'_'+params['target']+'_skyspec.pdf'))
   plt.close(1)

   # In addition, also make a plot showing all the sky locations

   # Start by assembling a white light image
   wl_im = np.nansum(data, axis = 0)

   # Very well, now let's create a fits file to save this as required.
   hdu0 = fits.PrimaryHDU(None, header0)
   # Add the white light image
   hdu1 = fits.ImageHDU(wl_im)
   # Make sure the WCS coordinates are included as well
   hdu1 = bifus_t.hdu_add_wcs(hdu1, header_data)
   # Also include a brief mention about which version of brutifus is being used
   hdu1 = bifus_t.hdu_add_brutifus(hdu1, suffix)

   # Write the file!
   hdu = fits.HDUList(hdus=[hdu0, hdu1])

   fn_list['wl_im'] = os.path.join(bifus_m.prod_loc,
                                   suffix+'_'+params['target']+'_wl-im.fits')
   hdu.writeto(fn_list['wl_im'], overwrite = True)

   # Image filename
   ofn = os.path.join(bifus_m.plot_loc,
                                   suffix+'_'+params['target']+'_sky-regions.pdf')

   # Make the base figure
   (fig, ax1, ofn) = bifus_p.make_2Dplot(fn_list['wl_im'], ext=1,
                                         ofn=ofn,
                                         stretch='linear',
                                         pmin=5,
                                         pmax=90)

   # Show all the sky spaxels
   #sky_spaxels[sky_spaxels ==0] = np.nan
   #ax1.imshow(sky_spaxels, origin= 'lower', cmap='winter', vmin = 0, vmax = 1, alpha=0.5)
   ax1.contour(sky_spaxels, levels=[0.5], colors = ['w'],
               linewidths = [0.75], origin = 'lower')
   # TODO: have the contours follow the pixel edges

   fig.savefig(ofn)

   # Now get started with the actual sky subtraction

   # Turn the spectrum into a cube
   sky_cube = np.repeat(np.repeat(sky_spec[:,np.newaxis],
                                  header_data['NAXIS2'], axis=1)[:,:,np.newaxis],
                        header_data['NAXIS1'], axis=2)

   # Perform the sky subtraction
   data -= sky_cube

   # Here, I assume that the sky correction is "perfect" (i.e. it comes with no errors)
   # Since it's just a subtraction, I don't need to alter the error cube

   # Export the new datacube
   hdu0 = fits.PrimaryHDU(None, header0)
   hdu1 = fits.ImageHDU(data)
   hdu2 = fits.ImageHDU(error)

   for hdu in [hdu1,hdu2]:
      # Make sure the WCS coordinates are included as well
      hdu = bifus_t.hdu_add_wcs(hdu, header_data)
      hdu = bifus_t.hdu_add_lams(hdu, header_data)
      # Also include a brief mention about which version of brutifus is being used
      hdu = bifus_t.hdu_add_brutifus(hdu, suffix)


   # Add the filename to the dictionary of filenames
   fn_list[name_out] = os.path.join(bifus_m.prod_loc,
                                    suffix+'_'+params['target']+'_skysub-cube.fits')
   hdu = fits.HDUList(hdus=[hdu0, hdu1, hdu2])
   hdu.writeto( fn_list[name_out], overwrite=True)


   return fn_list

# --------------------------------------------------------------------------------------------------
def run_gal_dered(fn_list, params, suffix = None, name_in = None, name_out = None):
   ''' Corrects for Galactic extinction, given the Ab and Av extinction.

   This function derives the Alambda value for any wavelength, and corrects the data to
   correct for our "local" extinction. Intended for extragalactic sources.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params: The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: str
   :param name_out: name tag to identify which cube comes out of the routine
   :type name_out: str

   :return: The updated dictionary of filenames.
   :rtype: dict

   .. note:: To reproduce the approach from NED, use the Ab and Av value for you object
             from there, and set ``curve='f99'``, ``rv=3.1`` in the params file.   
   '''

   if params['verbose']:
      print('-> Correcting for Galactic extinction.')

   # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Compute Alambda
   alams = bifus_red.alam(lams, params['Ab'],params['Av'], curve=params['gal_curve'],
                           rv=params['gal_rv'])

   # Compute the flux correction factor
   etau = bifus_red.galactic_red(lams,params['Ab'],params['Av'],
                                 curve=params['gal_curve'],
                                 rv=params['gal_rv'])

   # Correct the data and the propagate the errors
   data *= etau[:, np.newaxis, np.newaxis]
   error *= (etau[:, np.newaxis, np.newaxis])**2

   # Save the datacubes
   hdu0 = fits.PrimaryHDU(None,header0)
   hdu1 = fits.ImageHDU(data)
   hdu2 = fits.ImageHDU(error)

   for hdu in [hdu1,hdu2]:
      # Make sure the WCS coordinates are included as well
      hdu = bifus_t.hdu_add_wcs(hdu, header_data)
      hdu = bifus_t.hdu_add_lams(hdu, header_data)
      # Also include a brief mention about which version of brutifus is being used
      hdu = bifus_t.hdu_add_brutifus(hdu, suffix)
      # Add the line reference wavelength for future references
      hdu.header['BR_AB'] = (params['Ab'], 'Extinction in B')
      hdu.header['BR_AV'] = (params['Av'], 'Extinction in V')

   # Add the filename to the dictionary of filenames
   fn_list[name_out] = os.path.join(bifus_m.prod_loc,
                                    suffix+'_'+params['target']+'_gal-dered_cube.fits')

   hdu = fits.HDUList(hdus=[hdu0,hdu1,hdu2])
   hdu.writeto(fn_list[name_out], overwrite=True)

   # Make a plot of this.
   ofn = os.path.join(bifus_m.plot_loc,suffix+'_'+params['target']+
                            '_gal_Alambda_corr.pdf')
   bifus_p.make_galred_plot(lams,alams,etau,params['Ab'],params['Av'], ofn)

   return fn_list

# --------------------------------------------------------------------------------------------------
def run_fit_continuum(fn_list, params, suffix=None, name_in=None,
                      start_row=None, end_row=None,
                      method='lowess'):
   ''' Fits the stellar/nebular continuum using a specific method.

   This function fits the continuum in the datacube. It is designed to use multiprocessing
   to speed things up on good computers. It deals with the data columns-per-columns, and
   can be restarted mid-course, in case of a crash.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params: The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: str
   :param start_row: the row from which to sart the fitting. None for min.
   :type start_row: int
   :param end_row: the row on which to end the fitting (inclusive). None for max.
   :type end_row: int
   :param method: which method to use for the fitting ?
   :type method: str

   :return: The updated dictionary of filenames.
   :rtype: dict

   '''

   # For the bad continuum: run a LOWESS filter from statsmodel.nonparametric:
   # sm.nonparametric.smoothers_lowess.lowess(spec,lams,frac=0.05, it=5)

   # Rather than launch it all at once, let's be smart in case of problems. I'll run
   # the fits row-by-row with multiprocessing (hence up to 300cpus can help!), and save
   # between each row.

    # Get the data
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # I also need to load the SNR cube to know where I have data I want to fit
   '''
   ### DISABLED FOR NOW - ONLY MEANINGFUl WITH MORE THAN 1 CONT. FIT. TECHNIQUE ###
   try:
      print(os.path.join(bifus_m.prod_loc,fn_list['snr_maps']))
      hdu = fits.open(os.path.join(bifus_m.prod_loc,fn_list['snr_maps']))
      is_nonnan = hdu[1].data
      hdu.close()
   except:
      raise Exception('Failed to load the crude SNR cube. Did you run the step ?')
   '''

   # Get some info about the cube
   nrows = header_data['NAXIS1']

   # Where do I start and end the processing ?
   if start_row is None:
      start_row = 0
   if end_row is None:
      end_row = nrows-1

   # Ok, what do I want to do ?
   if method == 'lowess':
      if params['verbose']:
         print('-> Starting the continuum fitting using the LOWESS approach.')

      fit_func = partial(bifus_cof.lowess_fit, lams = lams,
                         frac = params['lowess_frac'], it = params['lowess_it'])
      # Note here the clever use of the partial function, that turns the lowess_fit
      # function from something that takes 4 arguments into something that only takes 1
      # argument ... thus perfect for the upcoming "map" functions !

   else:
      raise Exception('Continuum fitting method "%s" not supported.' % (method))

   # Very well, let's start the loop on rows. If the code crashes/is interrupted, you'll
   # loose the current row. Just live with it.
   for row in np.linspace(start_row, end_row, end_row - start_row + 1).astype(int):

      # Alright, now deal with the spaxels outside the user-chosen SNR range.
      # Replace them with nan's
      good_spaxels = np.ones((header_data['NAXIS2']))

      '''
      ### DISABLED FOR NOW - ONLY MEANINGFUl WITH MORE THAN 1 CONT. FIT. TECHNIQUE ###
      if params[method+'_snr_min']:
         good_spaxels[snr_cont[:,row] < params[method+'_snr_min']] = np.nan
      if params[method+'_snr_max']:
         good_spaxels[snr_cont[:,row] >= params[method+'_snr_max']] = np.nan
      '''

      # Build a list of spectra to be fitted
      specs = [data[:,i,row] * good_spaxels[i] for i in range(header_data['NAXIS2'])]
      errs = [error[:,i,row] * good_spaxels[i] for i in range(header_data['NAXIS2'])]

      # Set up the multiprocessing pool of workers
      if params['multiprocessing']:
         # Did the user specify a number of processes to use ?
         if type(params['multiprocessing']) == np.int:
            nproc = params['multiprocessing']

         else: # Ok, just use them all ...
            nproc = multiprocessing.cpu_count()

         pool = multiprocessing.Pool(processes = nproc,
                                     initializer = bifus_t.init_worker())

         if params['verbose']:
            sys.stdout.write('\r   Fitting spectra in row %2.i, %i at a time ...' %
                                 (row,nproc))
            sys.stdout.flush()

         # Launch the fitting ! Make sure I deal with KeyBoard Interrupt properly
         # Only a problem for multiprocessing. For the rest of the code, whatever.
         try:
            conts = pool.map(fit_func, specs)
         except KeyboardInterrupt:
            print(' interrupted !')
            # Still close and join properly
            pool.close()
            pool.join()
            sys.exit('Multiprocessing continuum fitting interrupted at row %i'% row)
         else: # If all is fine
            pool.close()
            pool.join()


      else: # just do things 1-by-1
         if params['verbose']:
            sys.stdout.write('\r   Fitting spectra in row %2.i, one at a time ...' %
                                 row)
            sys.stdout.flush()

         conts = map(fit_func, specs)

      # Here, I need to save these results. Pickle is fast and temporary,
      # until I re-build the entire cube later on. It also allows for better 
      # row-by-row flexibility.

      print(' ') # Need this to deal with the stdout mess

      if not(os.path.isdir(bifus_m.tmp_loc)):
         print('   !Requested storage location does not exist! Creating it ...')
         print('    => %s' % bifus_m.tmp_loc)
         os .mkdir(bifus_m.tmp_loc)

      fn = os.path.join(bifus_m.tmp_loc,
                        suffix+'_'+params['target']+'_'+method+'_row_'+
                        str(np.int(row)).zfill(4)+'.pkl')
      file = open(fn,'wb')
      pickle.dump(conts,file)
      file.close()

   # And add the generic pickle filename to the dictionary of filenames
   fn_list[method+'_pickle'] = suffix+'_'+params['target']+'_'+method+'_row_'

   print('   Fitting completed !')

   return fn_list

# ----------------------------------------------------------------------------------------
def run_make_continuum_cube(fn_list, params, suffix = None, method = 'lowess'):
   ''' Assemble a continuum cube from the individually-pickled fits.

   This function is designed to construct a "usable and decent" datacube out of the
   mess generated by the continuum fitting function, i.e. out of the many pickle
   files generated.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params: The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param method: which fits to gather ?
   :type method: str

   :return: The updated dictionary of filenames.
   :rtype: dict

   '''

   if params['verbose']:
      print('-> Constructing the datacube for the continuum fitting (%s).' % method)

   # Get the raw data open, to know how big things are ...
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list['raw_cube'],params['inst'])

   # Get some info
   nrows = header_data['NAXIS1']

   # Prepare a continuum cube structure
   cont_cube = np.zeros_like(data) * np.nan

   # Loop through the rows, and extract the results.
   # Try to loop through everything - in case this step was run in chunks.
   for row in range(0,nrows):
      progress = 100. * (row+1.)/nrows
      sys.stdout.write('\r   Building cube [%5.1f%s]' %
                         (progress,'%'))
      sys.stdout.flush()

      fn = os.path.join(bifus_m.tmp_loc,
                        fn_list[method+'_pickle']+str(np.int(row)).zfill(4)+'.pkl')

      if os.path.isfile(fn):
         # Very well, I have some fit here. Let's get them back
         myfile = open(fn,'rb')
         conts = pickle.load(myfile)
         myfile.close()

         # Mind the shape
         if method=='lowess':
            # I directly saved the continuum as an array just get it out.
            cont_cube[:,:,row] = np.array(conts).T

         else:
            raise Exception(' Continuum fitting method "%s" unknown.' % (method))

   # Very well, now let's create a fits file to save this as required.
   hdu0 = fits.PrimaryHDU(None,header0)
   hdu1 = fits.ImageHDU(cont_cube)
   if method == 'lowess':
      hdu2 = fits.ImageHDU(np.zeros_like(cont_cube)) # Keep all the errors to 0 for now.
   else:
      raise Exception('What errors do you have in your fitted sky ???')

   # Make sure the WCS coordinates are included as well
   for hdu in [hdu1,hdu2]:
      hdu = bifus_t.hdu_add_wcs(hdu,header_data)
      hdu = bifus_t.hdu_add_lams(hdu,header_data)

      # Also include a brief mention about which version of brutifus is being used
      hdu = bifus_t.hdu_add_brutifus(hdu,suffix)

   hdu = fits.HDUList(hdus=[hdu0, hdu1, hdu2])
   fn_list[method+'_cube'] = os.path.join(bifus_m.prod_loc,
                                          suffix+'_'+params['target']+'_'+method+'.fits')
   hdu.writeto(fn_list[method+'_cube'], overwrite=True)

   print(' ') # Required to clear the stdout stuff

   return fn_list

# --------------------------------------------------------------------------------------------------
def run_subtract_continuum(fn_list, params, suffix = None, name_in = None,
                           name_out = None, method = None):
   ''' Subtracts the sky from a given cube.

   :param fn_list: The dictionary containing all filenames created by brutifus.
   :type fn_list: dict
   :param params: The dictionary containing all paramaters set by the user.
   :type params: dict
   :param suffix: The tag of this step, to be used in all files generated for rapid id.
   :type suffix: str
   :param name_in: name tag to identify which cube to use to run the routine
   :type name_in: str
   :param name_out: name tag to identify which cube comes out of the routine
   :type name_out: str

   :return: The updated dictionary of filenames.
   :rtype: dict

   '''

   if params['verbose']:
      print('-> Subtracting the sky continuum (%s).' % method)

   # Get the data open
   [[lams, data, error], [header0, header_data, header_error]] = \
      bifus_t.extract_cube(fn_list[name_in],params['inst'])

   # Get the skycube open
   [[skylams, skydata, skyerror], [skyheader0, skyheader_data, skyheader_error]] = \
      bifus_t.extract_cube(fn_list[method+'_cube'],params['inst'])


   # Since this is a subtraction, and I assume no error on the sky, the errors remain unchanged
   data -= skydata

   # And save this to a new fits file
   hdu0 = fits.PrimaryHDU(None,header0)
   hdu1 = fits.ImageHDU(data)
   hdu2 = fits.ImageHDU(error)

   for hdu in [hdu1,hdu2]:
      # Make sure the WCS coordinates are included as well
      hdu = bifus_t.hdu_add_wcs(hdu,header_data)
      hdu = bifus_t.hdu_add_lams(hdu,header_data)
      # Also include a brief mention about which version of brutifus is being used
      hdu = bifus_t.hdu_add_brutifus(hdu,suffix)

   hdus = fits.HDUList(hdus=[hdu0,hdu1,hdu2])

   fn_list[name_out] = os.path.join(bifus_m.prod_loc,
                                    suffix+'_'+params['target']+'_'+method+
                                    '-contsub-cube.fits')
   hdus.writeto(fn_list[name_out], overwrite=True)

   return fn_list



# -*- coding: utf-8 -*-
'''
brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
Copyright (C) 2018-2019,  F.P.A. Vogt
 
-----------------------------------------------------------------------------------------
 
This file contains tools for the brutifus routines to create pretty plots seamlessly.

Created November 2018, F.P.A. Vogt - - frederic.vogt@alumni.anu.edu.au
'''
# ----------------------------------------------------------------------------------------

import os
import sys
import pickle

import numpy as np
import scipy.interpolate as interp
from scipy.ndimage.filters import gaussian_filter

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import matplotlib.gridspec as gridspec 
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

from astropy.stats import sigma_clipped_stats
from astropy.io import fits
import astropy.units as u
from astropy.wcs import WCS
import astropy.visualization as astrovis

from . import brutifus_metadata as bifus_m

# Set the proper plotting style
plt.style.use(bifus_m.plotstyle)

# ----------------------------------------------------------------------------------------
def crosshair(inner_r=1, pa = 0):
   ''' The path of an emtpy cross, useful for indicating targets without crowding the field.
   
   :params inner_r: empty inner radius. Default =1 (gap = tick length = 1/3 marker width
   :type inner_r: float
   :params pa: position angle (degrees)   
   :type pa: float
   
   :return: path
   
   '''
   
   verts = [(-1.5, 0),
            (-0.5*inner_r, 0),
            (0, 0.5*inner_r),
            (0, 1.5),
            (0.5*inner_r, 0),
            (1.5, 0),
            (0, -0.5*inner_r),
            (0, -1.5),
            (-1.5, 0),
            (-1.5, 0),
           ]
   
   pa = np.radians(pa)
   rot_mat = np.matrix([[np.cos(pa),-np.sin(pa)],[np.sin(pa),np.cos(pa)]])
   
   for (v, vert) in enumerate(verts):
      verts[v] = (vert*rot_mat).A[0] 
      
   codes = [mpl.path.Path.MOVETO,
            mpl.path.Path.LINETO,
            mpl.path.Path.MOVETO,
            mpl.path.Path.LINETO,
            mpl.path.Path.MOVETO,
            mpl.path.Path.LINETO,
            mpl.path.Path.MOVETO,
            mpl.path.Path.LINETO,
            mpl.path.Path.MOVETO,
            mpl.path.Path.CLOSEPOLY
            ]

   path = mpl.path.Path(verts, codes)
    
   return path


# ----------------------------------------------------------------------------------------
def get_fig_dims(nx, ny):
   ''' Returns all the necessary figure dimensions for gridspec, given the size of the 
       image to plot.
   
   :param nx: number of pixels along x axis
   :type nx: int
   :param ny: number of pixels along y axis
   :type ny: int
   
   :return: list containing the figure parameters [width, height, height_ratios, 
            width_ratios, left, right, botom, top, wspace, hspace]
   :rtype: list
   
   '''
   
    
   fig_height = bifus_m.margin_top + bifus_m.cb_thick + bifus_m.margin_bottom + \
                ny/nx * (bifus_m.fig_width - bifus_m.margin_right - bifus_m.margin_left)
   
   height_ratios = [bifus_m.cb_thick/fig_height, 1]
   
   left = bifus_m.margin_left/bifus_m.fig_width
   right = 1-bifus_m.margin_right/bifus_m.fig_width
   bottom = bifus_m.margin_bottom/fig_height
   top = 1-bifus_m.margin_top/fig_height
   
   wspace = 0.02
   hspace = 0.02
   
   return [np.round(bifus_m.fig_width,3), np.round(fig_height,3), 
           np.round(height_ratios,3), np.round([1],3), 
           np.round(left,3), np.round(right,3), 
           np.round(bottom,3), np.round(top,3), np.round(wspace,3), np.round(hspace,3)]

# ----------------------------------------------------------------------------------------
def get_im_stretch(stretch):
   ''' Returns a stretch to feed the ImageNormalize routine from Astropy.
   
   :param stretch: short, catchy name for the stretch I want. Possibilities are
                   'arcsinh' or 'linear'.
   :type stretch: string
   
   :return: A :class:`astropy.visualization.stretch` thingy ...
   
   '''
   
   if stretch == 'arcsinh':
      return astrovis.AsinhStretch()
   elif stretch == 'linear':
      return astrovis.LinearStretch()
   else:
      raise Exception('Ouch! Stretch %s unknown.' % (stretch))
      
# ----------------------------------------------------------------------------------------    
def get_im_interval(pmin = 10, pmax = 99.9, vmin = None, vmax = None):
   ''' Returns an interval, to feed the ImageNormalize routine from Astropy.
   
   :param pmin: lower-limit percentile
   :type pmin: float
   :param pmax: upper-limit percentile
   :type pmax: float
   :param vmin: absolute lower limit
   :type vmin: float
   :param vmax: absolute upper limit
   :type vmax: float
   
   :return: an :class:`astropy.visualization.interval` thingy ...
   
   .. note:: Specifying *both* vmin and vmax will override pmin and pmax.
   
   '''
   
   if not(vmin is None) and not (vmax is None):
      
      return astrovis.ManualInterval(vmin,vmax)
   
   else:
      
      return astrovis.AsymmetricPercentileInterval(pmin,pmax)
      
# ----------------------------------------------------------------------------------------    
def finetune_WCSAxes(im_ax):
   ''' Fine-tune the look of WCSAxes plots to my liking.
   
   Currently, this changes the tick label format to show integer seconds, sets the 
   separators to hms/d ' ", enables minor ticks, and tweaks the axis labels to be R.A. and 
   Dec. (with dots!). 
   
   :param im_ax: the axes to be improved
   :type im_ax: :class:`matplotlib.axes`
   
   :return: (ra,dec), the axes coordinates.
   
   ''' 
   
   im_ax.set_facecolor((0.5,0.5,0.5))
   
   # Deal with the axes
   ra = im_ax.coords[0]
   dec = im_ax.coords[1]
   
   ra.set_major_formatter('hh:mm:ss')
   dec.set_major_formatter('dd:mm:ss')
   
   # This only works when I am using proper LaTeX
   ra.set_separator((r'$^\mathrm{h}$', r'$^\mathrm{m}$', r'$^\mathrm{s}$ '))
   dec.set_separator((r'$^{\circ}$', r'$^{\prime}$', r'$^{\prime\prime}$'))
   
   ra.display_minor_ticks(True)
   dec.display_minor_ticks(True)
   
   im_ax.set_xlabel('R.A. [J2000]')
   im_ax.set_ylabel('Dec. [J2000]', labelpad=0)
   
   return (ra,dec)

# ----------------------------------------------------------------------------------------  
# Here, I define the "alligator" colormap to give brutifus a unique feel, just because I 
# can. More importantly, I can decide to set the lower and upper bounds to B&W, so that 
# spaxels outside the colorbar range are directly identifiable. Also, I can make sure that 
# nan's show up in 50% grey. 

# Alligator: 6 nodes with linear evolution, suitable for B&W impression
cdict_alligator = {
'red'  :  ( (0.00, 0./255, 0./255),
            (0.00, 0./255, 0./255),      (0.2, 20./255., 20./255.),
            (0.40, 46./255., 46./255.),  (0.6, 108./255., 108./255.),
            (0.8, 207./255., 207./255.), (1.00, 255./255.,255./255.),
            (1.00, 255./255, 255./255),),

'green' : ( (0.00, 0./255, 0./255),
            (0.00, 25./255., 25./255.),  (0.2, 85./255., 85./255.),
            (0.4, 139./255., 139./255.), (0.6, 177./255., 177./255.),  
            (0.8, 234./255, 234./255),   (1.00, 248./255.,248./255.),
            (1.00, 255./255, 255./255),),
            
'blue':   ( (0.00, 0./255, 0./255),
            (0.00, 25./255., 25./255.),  (0.2, 81./255., 81./255.),
            (0.4, 87./255., 87./255.),   (0.6, 86./255., 86./255.),
            (0.8, 45./255, 45./255),     (1.00, 215./255.,215./255.),
            (1.00, 255./255, 255./255),),   
}

# Initiate the colorbar
alligator = plt.matplotlib.colors.LinearSegmentedColormap('alligator',cdict_alligator, 1024)
# Set the bad colors
alligator.set_bad(color=(0.5,0.5,0.5), alpha=1) 

# ----------------------------------------------------------------------------------------
def reverse_colourmap(cdict):
   ''' Returns the inverted colorbar dictionary.

   I most definitely got this from the web somewhere ... Stackoverflow?
      
   :params cdict: a colorbar dictionary
   :type cdict: dict
      
   :return: the flipped dictionary
   :rtype: dict
      
   '''

   new_cdict = {}
   for channel in cdict:
      data = []
      for t in cdict[channel]:
         data.append((1. - t[0], t[1], t[2]))
        
      new_cdict[channel] = data[::-1]
   return new_cdict

# ----------------------------------------------------------------------------------------
# Define the inverse alligator cmap as well
cdict_alligator_r = reverse_colourmap(cdict_alligator)
alligator_r = plt.matplotlib.colors.LinearSegmentedColormap('alligator_r', 
                                                            cdict_alligator_r, 1024) 
# Set the bad colors
alligator_r.set_bad(color=(0.5,0.5,0.5), alpha=1) 

# ----------------------------------------------------------------------------------------      
def make_galred_plot(lams, alams, etau, Ab, Av, ofn):
   ''' Plots the extinction Alambda and flux correction factor for galactic extinction.
    
   :param lams: The wavelength nodes.
   :type lams: 1D numpy array
   :param alams: The corresponding values of Alambda.
   :type lams: 1D numpy array
   :param etau: The flux correction factor, i.e. F_(unextinct)/F_(observed).
   :type lams: 1D numpy array
   :param Ab: The value of Ab.
   :type lams: float
   :param Av: The value of Av.
   :type lams: float
   :param ofn: path+name of where to save the plot
   :type ofn: string 
   
   '''
   
   # Start the plotting
   plt.close(1)
   plt.figure(1,figsize=(6.93,6))
   gs = gridspec.GridSpec(1,1, height_ratios=[1], width_ratios=[1])
   gs.update(left=0.14, right=0.85, bottom=0.12, top=0.93, wspace=0.02, hspace=0.03)
  
   ax = plt.subplot(gs[0,0])
   
   ax.plot(lams,alams,'k-',drawstyle='steps-mid')
   
   ax.set_xlabel(r'Wavelength [\AA]')
   ax.set_ylabel(r'A$_\lambda$ [mag]')
   
   ax2 = ax.twinx()
   ax2.plot(lams,etau,'--',c='firebrick',drawstyle='steps-mid')
   ax2.set_ylabel(r'F$_{\lambda}$/F$_{\lambda,obs}= e^{\tau}$', color='firebrick',
                  labelpad=10)
   
   # make the other y-axis crimson for clarity
   ax2.tick_params(axis='y',colors='firebrick', which='both')
   ax2.spines['right'].set_color('firebrick')
   
   # Add the legend
   ax.text(0.8,0.9,'A$_{B}$=%.3f\n A$_{V}$=%.3f' % (Ab,Av), transform=ax.transAxes,
           verticalalignment='center',
           horizontalalignment='center',
         )
   
   # Tweak the limits
   ax.set_xlim((np.min(lams),np.max(lams)))
   
   plt.savefig(ofn)
   plt.close() 
   
# ----------------------------------------------------------------------------------------         
def show_scale(ax, scale_length = 10.*u.arcsec, scale_text = None, scale_loc = 'lower left'):
   ''' Adds a scalebar to a given 2D plot. 
   
   :param ax: The axes to which to add the sale.
   :param scale_length: The scale length in astropy.units
   :param scale_text: the string to print as the scalebar, or None
   :param scale_loc: string [default: 'top left']
                 The location of the scale bar. e.g 'top right', 'bottom left', etc ...
                 
   '''
   
   raise Exception('Ouch! The scalebar is not working yet ...!')
   
   # Make a default scale text in case none is specified
   if scale_text is None:
      
      the_unit = scale_length.unit.to_string()
      
      if the_unit == 'arcsec':
         the_unit = r'$^{\prime\prime}$'
      elif the_unit == 'arcmin':
         the_unit = r'$^{\prime}$'
      else:
         the_unit = ' ' + the_unit
                
      scale_text = '%.1f%s' % (scale_length.value, the_unit)
   
   # TODO:
   # Looks like I still need to divide the length by the "pixel scale" ? 
   # Unless I need to divide by the Dec ?
   # But how do I extract this from the ax ???
      
   # Create the scalebar
   bar = AnchoredSizeBar(ax.get_transform('world'), scale_length.to(u.degree).value, 
                         scale_text, scale_loc, 
                         sep = 5, pad = 0.5, borderpad = 0.4)
    
   # Adjust the bar thickness
   bar.size_bar.get_children()[0].set_linewidth(2)
   # Add it to the plot
   ax.add_artist(bar)
   
# ----------------------------------------------------------------------------------------      

def make_2Dplot(fn, # path to the data (complete!)
                ext = 0, # Which extension am I looking for ?
                ofn = 'plot.pdf', # Savefig filename
                stretch = 'arcsinh',
                vlims = [None, None], 
                plims = [10.0, 99.99],
                gauss_blur = None,
                cmap = None,
                cblabel = None,
                scalebar = None,
                ):
   ''' Creates an image from a 2-D fits image with WCS info.

   :param fn: The filename (+path!) fo the fits file to display.
   :type fn: string
   :param ext: The extension of the image to display. The data in this extension MUST be 2-D.
   :type ext: int
   :param ofn: The output filename (+path!)
   :type ofn: string
   :param stretch: The stretch of the image, fed to aplpy.FITSFigure. 'linear', 'arcsinh', ...
   :type stretch: string
   :param vmin: If set, the lower bound of the colorbar
   :type vmin: float
   :param vmax: If set, the upper bound of the colorbar
   :type vmax: float
   :param pmin: If set, the lower percentile bound of the colorbar
   :type pmin: float
   :param pmax: If set, the upper percentile bound of the colorbar
   :type pmax: float
   :param cmap: If set, the colormap to use. Use 'alligator' for the special brutifus cmap.
   :type cmap: string
   :param cblabel: If set, the label of the colorbar.
   :type cblabel: string
   :param scalebar: If set, adds a scale bar to the plot. Format: [lenght arcsec, length kpc, loc, unit]
   :type scalebar: list of len(3)
            
   :return: a list containing the figure, the ax1 and the plot filename.
   :rtype: list
        
   .. note:: This function absolutely requires WCS coordinates, and a 2-D array.
      
   '''
   
   # Alright, let's start by computing the proper figure height, so that my plots all
   # have the same ratio, irrespective of whether I have a colorbar, or not.
   
   # First get the data
   
   hdu = fits.open(fn)
   data = hdu[ext].data
   header = hdu[ext].header
   hdu.close()
   
   # What is the size of my image ?
   (ny,nx) = np.shape(data)
   
   # What is the associated fig dimensions I need ?
   fig_dims = get_fig_dims(nx, ny)
   
   # If requested, smooth the array
   if not(gauss_blur is None):
      data = gaussian_filter(data, sigma = gauss_blur)
      
   # Let's create a plot, and forget about using aplpy
   plt.close(1)
   fig = plt.figure(1, figsize=(fig_dims[0], fig_dims[1]))
   
   # Set the scene with gridspec
   gs = gridspec.GridSpec(2, 1, height_ratios=fig_dims[2], width_ratios=fig_dims[3],
                          left=fig_dims[4], right=fig_dims[5], bottom=fig_dims[6], top=fig_dims[7],
                          wspace=fig_dims[8], hspace=fig_dims[9])
   
   ax1 = plt.subplot(gs[1, 0], projection=WCS(header))
   
   # What is the colorscheme selected ?
   if cmap == 'alligator':
      mycmap = alligator
   elif cmap == 'alligator_r':
      mycmap = alligator_r
   elif cmap is None:
      mycmap = 'Greys'
   else:
      mycmap = cmap

   # What is the normalization ?
   norm = astrovis.ImageNormalize(data,
                                  interval=get_im_interval(pmin=plims[0], pmax=plims[1],
                                                           vmin=vlims[0], vmax=vlims[1]),
                                  stretch=get_im_stretch(stretch),
                                  clip=False)

   # Plot the image
   im = ax1.imshow(data, origin='lower', interpolation='nearest', cmap=mycmap,
                   norm=norm)

   # This does not work !!!!
   # TODO:
   ax1.set_facecolor('salmon')

   # Add the scalebar
   #if scalebar is not None:
   #   show_scale(ax1, scale_length = scalebar[0], scale_text = scalebar[1],
   #              scale_loc = scalebar[2])

   # Make the axes look pretty
   (ra,dec) = finetune_WCSAxes(ax1)

   # Deal with the colorbar
   if cmap is not None:

      ax0 = plt.subplot(gs[0,0])

      cb = plt.colorbar(cax = ax0, mappable=im, orientation='horizontal',
                        ticklocation='top')
      cb.set_label(cblabel, labelpad = 10)
      #cb.ax.xaxis.set_ticks_position('top')
      ax0.tick_params(axis='x', which='major', pad = 5)

      # BUG: can't get the minor ticks working properly - turn them off for now
      cb.ax.minorticks_off()

   fig.savefig(ofn)

   return (fig, ax1, ofn)

# --------------------------------------------------------------------------------------------------
def make_RGBplot(fns, ofn, ext = [0, 0, 0],
                 stretch = ['linear', 'linear', 'linear'],
                 plims = [0.25, 99.75, 0.25, 99.75, 0.25, 99.75],
                 vlims = [None, None, None, None, None, None],
                 gauss_blur = [None, None, None],
                 title = None,
                 scalebar = None,
                ):
   ''' Creates an RGB image from three fits files.

   :param fns: The filename (+path!) fo the  3 fits file to display (in R, G and B orders).
   :type fns: list
   :param ofn: The filneame (+path) of the output file.
   :type ofn: str
   :param ext: What HDU extension to read ?
   :type ext: list
   :param stretch: The stretch to apply to the data, e.g. 'linear', 'log', 'arcsinh'.
   :type stretch: list
   :param plims: The limiting percentiles for the plot, as 
                 [pmin_r, pmax_r, pmin_g, pmax_g, pmin_b, pmax_b]
   :type plims: list
   :param vlims: The limting values for the plot (superseeds plims), as
                 [vmin_r, vmax_r, vmin_g, vmax_g, vmin_b, vmax_b]
   :type vlims: list
   :param gauss_blur: list of radius (in pixels) for gaussian_blur. None = no blur
   :type gauss_blur: list
   :param title: Title to display above the image
   :type title: str
   :param scalebar: If set, adds a scale bar to the plot.
                    Format: [lenght arcsec, length kpc, loc, unit]
   :type scalebar: list 

   :return: a list containing the figure, the ax1 and the plot filename.
   :rtype: list
   
   '''

   # If I was given a single stretch for all images, enlarge it
   if type(stretch) == np.str:
      stretch = [stretch, stretch, stretch]

   # If I was given a single extension, assume it is the same everywhere
   if type(ext) == np.int:
      ext = [ext, ext, ext]

   # Open the data and headers
   data = []
   header = []

   for (f, fn) in enumerate(fns):
      hdu = fits.open(fn)

      # Get the data
      this_data = hdu[ext[f]].data

      # If requested, smooth the array
      if gauss_blur[f] is not None:

         this_data = gaussian_filter(this_data, sigma=gauss_blur[f])

      # If the image if full of NaN's ... issue a proper error, 'cause I can't normalize it!
      if np.size(this_data[this_data == this_data]) == 0:
         raise Exception('Ouch ... the %s image if full of NaNs!' % (['R', 'G ', 'B'][f]))

      # Normalize it
      this_data = get_im_interval(pmin=plims[2 * f], pmax=plims[2 * f + 1],
                                  vmin=vlims[2 * f], vmax=vlims[2 * f + 1])(this_data)

      # Stretch it
      this_data = get_im_stretch(stretch[f])(this_data)

      data += [this_data]
      header += [hdu[ext[f]].header]

    # What is the size of my images ?
   (ny,nx) = np.shape(data[0])

   # What is the associated fig dimensions I need ?
   fig_dims = get_fig_dims(nx, ny)

   # Let's create a plot, and forget about using aplpy
   plt.close(1)
   fig = plt.figure(1, figsize=(fig_dims[0], fig_dims[1]))

   # Set the scene with gridspec
   gs = gridspec.GridSpec(2, 1, height_ratios=fig_dims[2], width_ratios=fig_dims[3],
                          left=fig_dims[4], right=fig_dims[5], bottom=fig_dims[6], top=fig_dims[7],
                          wspace=fig_dims[8], hspace=fig_dims[9])

   ax1 = plt.subplot(gs[1, 0], projection=WCS(header[0]))

   im = ax1.imshow(np.transpose(np.array(data), (1, 2, 0)), origin='lower', interpolation='nearest', 
                   zorder=0)

   # Assume an astro image, and set the ticks to white
   ax1.tick_params(axis='both', which='both', color='w', direction='in')

   # Make the axes look pretty
   (ra,dec) = finetune_WCSAxes(ax1)

   if title is not None:
      ax1.set_title(title)

   # Save the figure
   fig.savefig(ofn)

   return (fig, ax1, ofn)



# -*- coding: utf-8 -*-
'''
 brutifus: a set of Python modules to process datacubes from integral field spectrographs.\n
 Copyright (C) 2018-2019,  F.P.A. Vogt
 
 -----------------------------------------------------------------------------------------
 
 This file contains functions related to the continuum fitting inside brutifus. 

 Created November 2018, F.P.A. Vogt - frederic.vogt@alumni.anu.edu.au
'''
# ----------------------------------------------------------------------------------------

import numpy as np
from statsmodels.nonparametric.smoothers_lowess import lowess

# ----------------------------------------------------------------------------------------

def lowess_fit(spec, lams, frac=0.05, it=5):
   ''' Fit a spectrum using a Locally Weighted Scatterplot Smoothing approach.
    
   Wraps around statsmodels.nonparametric.smoothers_lowess.lowess().
    
   :param spec: 1-D numpy array
            The input spectrum.
   :param lams: 1-D numpy array
            The corresponding wavelength array.
   :param frac: float [default:0.05]
            Between 0 and 1. The fraction of the data used when estimating each y-value.
            [From the statsmodel lowess function]
   :param it: int [default:5]
         The number of residual-based reweightings to perform.
         [From the statsmodel lowess function]
     
   :return: The fitted array, with size equal to spec.  
   :rtype: numpy array 
    
   .. note:: This function fits a spectrum using a LOWESS (Locally Weighted Scatterplot 
             Smoothing) technique, described in: 
             Cleveland, W.S. (1979) Robust Locally Weighted Regression and Smoothing 
             Scatterplots. Journal of the American Statistical Association 74 (368): 829-836.
    
             This is robust to outliers (hot pixels, cosmics), and is also efficient to 
             ignore emission lines. frac=0.05 and it=5 seem to work very fine for spectra 
             of any SNR, both lousy with no continuum, and good ones in the center of 
             galaxies - modulo the stellar absorption features which are of course 
             "ignored" by the LOWESS routine.
   
   ..warning:: If you have broad emission line in the spectrum, you want to be *very*
               careful with LOWESS !!!
               
   '''
    
   # Only do the fit if there is some signal. Avoid an ugly warning in the prompt.
   if np.all(np.isnan(spec)):
      fit = np.zeros_like(spec) * np.nan
   else:
      fit = lowess(spec,lams,frac=frac, it=it, is_sorted=True, missing = 'drop', 
                   return_sorted=False)
                                            
   return fit
   
# ----------------------------------------------------------------------------------------      
